/*
 $License:
    Copyright (C) 2010 InvenSense Corporation, All Rights Reserved.
 $
 */
/**
 *  @defgroup   ACCELDL (Motion Library - Accelerometer Driver Layer)
 *  @brief      Provides the interface to setup and handle an accelerometers
 *              connected to the secondary I2C interface of the gyroscope.
 *
 *  @{
 *      @file   lis331.c
 *      @brief  Accelerometer setup and handling methods for Invensense MANTIS
 */

/* ------------------ */
/* - Include Files. - */
/* ------------------ */

#ifdef __KERNEL__
#include <linux/module.h>
#endif

#include "mpu.h"
#include "mlsl.h"
#include "mlos.h"

#include <log.h>
#undef MPL_LOG_TAG
#define MPL_LOG_TAG "MPL-acc"

/* --------------------- */
/* -    Variables.     - */
/* --------------------- */

struct mantis_config {
	unsigned int odr; /* output data rate 1/1000 Hz*/
	unsigned int fsr; /* full scale range mg */
	unsigned int ths; /* Motion no-motion thseshold mg */
	unsigned int dur; /* Motion no-motion duration ms */
};

struct mantis_private_data {
	struct mantis_config suspend;
	struct mantis_config resume;
};


/*****************************************
 *Accelerometer Initialization Functions
 *****************************************/
/**
 * Record the odr for use in computing duration values.
 *
 * @param config Config to set, suspend or resume structure
 * @param odr output data rate in 1/1000 hz
 */
static int mantis_set_odr(void *mlsl_handle,
			struct ext_slave_platform_data *slave,
			struct mantis_config *config,
			long apply,
			long odr)
{
	config->odr = odr;
	return ML_SUCCESS;
}

static int mantis_set_ths(void *mlsl_handle,
			struct ext_slave_platform_data *slave,
			struct mantis_config *config,
			long apply,
			long ths)
{
	if (ths < 0)
		ths = 0;

	config->ths = ths;
	MPL_LOGV("THS: %d\n", config->ths);
	return ML_SUCCESS;
}

static int mantis_set_dur(void *mlsl_handle,
			struct ext_slave_platform_data *slave,
			struct mantis_config *config,
			long apply,
			long dur)
{
	if (dur < 0)
		dur = 0;

	config->dur = dur;
	MPL_LOGV("DUR: %d\n", config->dur);
	return ML_SUCCESS;
}

static int mantis_set_fsr(void *mlsl_handle,
			struct ext_slave_platform_data *slave,
			struct mantis_config *config,
			long apply,
			long fsr)
{
	if (fsr <= 2000)
		config->fsr = 2000;
	else if (fsr <= 4000)
		config->fsr = 4000;
	else
		config->fsr = 8000;

	MPL_LOGV("FSR: %d\n", config->fsr);
	return ML_SUCCESS;
}

#ifdef UMPL
static struct mantis_private_data mantis_private_data_instance;
#endif

static int mantis_init(void *mlsl_handle,
		       struct ext_slave_descr *slave,
		       struct ext_slave_platform_data *pdata)
{
	struct mantis_private_data *private_data;
#ifndef UMPL
	private_data = (struct mantis_private_data *)
		MLOSMalloc(sizeof(struct mantis_private_data));
#else
    private_data = &mantis_private_data_instance;
#endif

	if (!private_data)
		return ML_ERROR_MEMORY_EXAUSTED;

	pdata->private_data = private_data;

	mantis_set_odr(mlsl_handle, pdata, &private_data->suspend, FALSE, 0);
	mantis_set_odr(mlsl_handle, pdata,
		&private_data->resume, FALSE, 200000);
	mantis_set_fsr(mlsl_handle, pdata, &private_data->suspend, FALSE, 2000);
	mantis_set_fsr(mlsl_handle, pdata, &private_data->resume, FALSE, 2000);
	mantis_set_ths(mlsl_handle, pdata, &private_data->suspend, FALSE, 80);
	mantis_set_ths(mlsl_handle, pdata, &private_data->resume, FALSE, 40);
	mantis_set_dur(mlsl_handle, pdata, &private_data->suspend, FALSE, 1000);
	mantis_set_dur(mlsl_handle, pdata, &private_data->resume, FALSE,  2540);
	return ML_SUCCESS;
}

static int mantis_exit(void *mlsl_handle,
			  struct ext_slave_descr *slave,
			  struct ext_slave_platform_data *pdata)
{
#ifndef UMPL	
	if (pdata->private_data)
		return MLOSFree(pdata->private_data);
	else
		return ML_SUCCESS;
#else
	return ML_SUCCESS;
#endif		
}

int mantis_suspend(void *mlsl_handle,
		   struct ext_slave_descr *slave,
		   struct ext_slave_platform_data *pdata)
{
	unsigned char reg;
	int result;

	result = MLSLSerialRead(mlsl_handle, pdata->address,
				MPUREG_PWR_MGMT_2, 1, &reg);
	ERROR_CHECK(result);
	reg |= (BIT_STBY_XA | BIT_STBY_YA | BIT_STBY_ZA);

	result = MLSLSerialWriteSingle(mlsl_handle, pdata->address,
				MPUREG_PWR_MGMT_2, reg);
	ERROR_CHECK(result);

	return ML_SUCCESS;
}

int mantis_resume(void *mlsl_handle,
		  struct ext_slave_descr *slave,
		  struct ext_slave_platform_data *pdata)
{
	int result = ML_SUCCESS;
	unsigned char reg;
	struct mantis_private_data *private_data;

	private_data = (struct mantis_private_data *) pdata->private_data;

#ifdef MPU6000B
	MLSLSerialRead(mlsl_handle, pdata->address,
		MPUREG_PWR_MGMT_1, 1, &reg);
	if (reg & 0x40) {
		MLSLSerialWriteSingle(mlsl_handle, pdata->address,
				MPUREG_PWR_MGMT_1, 1, 1reg & ~0x40);
		MLOSSleep(2);
	}
#endif
	MLSLSerialRead(mlsl_handle, pdata->address,
		MPUREG_PWR_MGMT_2, 1, &reg);

	reg &= ~(BIT_STBY_XA | BIT_STBY_YA | BIT_STBY_ZA);

	result = MLSLSerialWriteSingle(mlsl_handle, pdata->address,
				MPUREG_PWR_MGMT_2, reg);
	ERROR_CHECK(result);

	if (slave->range.mantissa == 2)
		reg = 0;
	else if (slave->range.mantissa == 4)
		reg = 1 << 3;
	else if (slave->range.mantissa == 8)
		reg = 2 << 3;
	else if (slave->range.mantissa == 16)
		reg = 3 << 3;
	else
		return ML_ERROR;

	result = MLSLSerialWriteSingle(mlsl_handle, pdata->address,
				       MPUREG_ACCEL_CONFIG, reg);
	ERROR_CHECK(result);

	reg = (unsigned char) private_data->suspend.ths / ACCEL_MOT_THR_LSB;
	result = MLSLSerialWriteSingle(mlsl_handle, pdata->address,
				MPUREG_ACCEL_MOT_THR, reg);
	ERROR_CHECK(result);

	reg = (unsigned char)
		ACCEL_ZRMOT_THR_LSB_CONVERSION(private_data->resume.ths);
	result = MLSLSerialWriteSingle(mlsl_handle, pdata->address,
				MPUREG_ACCEL_ZRMOT_THR, reg);
	ERROR_CHECK(result);

	reg = (unsigned char) private_data->suspend.ths / ACCEL_MOT_DUR_LSB;
	result = MLSLSerialWriteSingle(mlsl_handle, pdata->address,
				MPUREG_ACCEL_MOT_DUR, reg);
	ERROR_CHECK(result);

	reg = (unsigned char) private_data->resume.ths / ACCEL_ZRMOT_DUR_LSB;
	result = MLSLSerialWriteSingle(mlsl_handle, pdata->address,
				MPUREG_ACCEL_ZRMOT_DUR, reg);
	ERROR_CHECK(result);
	return result;
}

int mantis_read(void *mlsl_handle,
		struct ext_slave_descr *slave,
		struct ext_slave_platform_data *pdata, unsigned char *data)
{
	int result;
	result = MLSLSerialRead(mlsl_handle, pdata->address,
				slave->read_reg, slave->read_len, data);
	return result;
}

static int mantis_config(void *mlsl_handle,
			struct ext_slave_descr *slave,
			struct ext_slave_platform_data *pdata,
			struct ext_slave_config *data)
{
	struct mantis_private_data *private_data = pdata->private_data;
	if (!data->data)
		return ML_ERROR_INVALID_PARAMETER;

	switch (data->key) {
	case MPU_SLAVE_CONFIG_ODR_SUSPEND:
		return mantis_set_odr(mlsl_handle, pdata,
				&private_data->suspend,
				data->apply,
				*((long *)data->data));
	case MPU_SLAVE_CONFIG_ODR_RESUME:
		return mantis_set_odr(mlsl_handle, pdata,
				&private_data->resume,
				data->apply,
				*((long *)data->data));
	case MPU_SLAVE_CONFIG_FSR_SUSPEND:
		return mantis_set_fsr(mlsl_handle, pdata,
				&private_data->suspend,
				data->apply,
				*((long *)data->data));
	case MPU_SLAVE_CONFIG_FSR_RESUME:
		return mantis_set_fsr(mlsl_handle, pdata,
				&private_data->resume,
				data->apply,
				*((long *)data->data));
	case MPU_SLAVE_CONFIG_MOT_THS:
		return mantis_set_ths(mlsl_handle, pdata,
				&private_data->suspend,
				data->apply,
				*((long *)data->data));
	case MPU_SLAVE_CONFIG_NMOT_THS:
		return mantis_set_ths(mlsl_handle, pdata,
				&private_data->resume,
				data->apply,
				*((long *)data->data));
	case MPU_SLAVE_CONFIG_MOT_DUR:
		return mantis_set_dur(mlsl_handle, pdata,
				&private_data->suspend,
				data->apply,
				*((long *)data->data));
	case MPU_SLAVE_CONFIG_NMOT_DUR:
		return mantis_set_dur(mlsl_handle, pdata,
				&private_data->resume,
				data->apply,
				*((long *)data->data));
	case MPU_SLAVE_CONFIG_IRQ_SUSPEND:
#if 0
		return mantis_set_irq(mlsl_handle, pdata,
				&private_data->suspend,
				data->apply,
				*((long *)data->data));
#endif
		break;
	case MPU_SLAVE_CONFIG_IRQ_RESUME:
#if 0
		return mantis_set_irq(mlsl_handle, pdata,
				&private_data->resume,
				data->apply,
				*((long *)data->data));
#endif
		break;
	default:
		return ML_ERROR_FEATURE_NOT_IMPLEMENTED;
	};

	return ML_SUCCESS;
}

static int mantis_get_config(void *mlsl_handle,
			struct ext_slave_descr *slave,
			struct ext_slave_platform_data *pdata,
			struct ext_slave_config *data)
{
	struct mantis_private_data *private_data = pdata->private_data;
	if (!data->data)
		return ML_ERROR_INVALID_PARAMETER;

	switch (data->key) {
	case MPU_SLAVE_CONFIG_ODR_SUSPEND:
		(*(unsigned long *)data->data) =
			(unsigned long) private_data->suspend.odr;
		break;
	case MPU_SLAVE_CONFIG_ODR_RESUME:
		(*(unsigned long *)data->data) =
			(unsigned long) private_data->resume.odr;
		break;
	case MPU_SLAVE_CONFIG_FSR_SUSPEND:
		(*(unsigned long *)data->data) =
			(unsigned long) private_data->suspend.fsr;
		break;
	case MPU_SLAVE_CONFIG_FSR_RESUME:
		(*(unsigned long *)data->data) =
			(unsigned long) private_data->resume.fsr;
		break;
	case MPU_SLAVE_CONFIG_MOT_THS:
		(*(unsigned long *)data->data) =
			(unsigned long) private_data->suspend.ths;
		break;
	case MPU_SLAVE_CONFIG_NMOT_THS:
		(*(unsigned long *)data->data) =
			(unsigned long) private_data->resume.ths;
		break;
	case MPU_SLAVE_CONFIG_MOT_DUR:
		(*(unsigned long *)data->data) =
			(unsigned long) private_data->suspend.dur;
		break;
	case MPU_SLAVE_CONFIG_NMOT_DUR:
		(*(unsigned long *)data->data) =
			(unsigned long) private_data->resume.dur;
		break;
	case MPU_SLAVE_CONFIG_IRQ_SUSPEND:
#if 0
		(*(unsigned long *)data->data) =
			(unsigned long) private_data->suspend.irq_type;
#endif
		break;
	case MPU_SLAVE_CONFIG_IRQ_RESUME:
#if 0
		(*(unsigned long *)data->data) =
			(unsigned long) private_data->resume.irq_type;
#endif
		break;
	default:
		return ML_ERROR_FEATURE_NOT_IMPLEMENTED;
	};

	return ML_SUCCESS;
}

struct ext_slave_descr mantis_descr = {
	/*.init             = */ mantis_init,
	/*.exit             = */ mantis_exit,
	/*.suspend          = */ mantis_suspend,
	/*.resume           = */ mantis_resume,
	/*.read             = */ mantis_read,
	/*.config           = */ mantis_config,
	/*.get_config       = */ mantis_get_config,
	/*.name             = */ "mantis",
	/*.type             = */ EXT_SLAVE_TYPE_ACCELEROMETER,
	/*.id               = */ ACCEL_ID_MPU6050,
	/*.reg              = */ 0x3B,
	/*.len              = */ 6,
	/*.endian           = */ EXT_SLAVE_BIG_ENDIAN,
	/*.range            = */ {2, 0},
	/*.trigger          = */ NULL,
};

struct ext_slave_descr *mantis_get_slave_descr(void)
{
	return &mantis_descr;
}
EXPORT_SYMBOL(mantis_get_slave_descr);

/**
 *  @}
 */

