/*
 $License:
    Copyright (C) 2010 InvenSense Corporation, All Rights Reserved.
 $
 */
/******************************************************************************
 *
 * $Id: mlarray.c 5085 2011-04-08 22:25:14Z phickey $
 *
 *****************************************************************************/

/**
 *  @defgroup ML
 *  @{
 *      @file   mlarray.c
 *      @brief  APIs to read different data sets from FIFO.
 */

/* ------------------ */
/* - Include Files. - */
/* ------------------ */
#include "ml.h"
#include "mltypes.h"
#include "mlinclude.h"
#include "mlMathFunc.h"
#include "mlmath.h"
#include "mlstates.h"
#include "mlFIFO.h"
#include "mlsupervisor.h"
#include "mldl.h"
#include "dmpKey.h"
#include "compass.h"

/**
 *  @brief  MLGetGyrosFixed is used to get gyro data from FIFO.
 *
 *
 *  @pre    MLDmpOpen() \ifnot UMPL or MLDmpPedometerStandAloneOpen() \endif
 *          must have been called.
 *
 *  @param  data
 *              A pointer to an array to be passed back to the user.
 *              <b>Must be 3 cells long </b>.
 *
 *  @return Zero if the command is successful; an ML error code otherwise.
 */
tMLError MLGetGyrosFixed(long * data)
{
    tMLError result = ML_SUCCESS;
    if ( MLGetState() < ML_STATE_DMP_OPENED )
        return ML_ERROR_SM_IMPROPER_STATE;

    if (NULL == data) {
        return ML_ERROR_INVALID_PARAMETER;
    }
            result = FIFOGetGyro( data );

    return result;
}

/**
 *  @brief  MLGetAccelsFixed is used to get accel data from FIFO.
 *
 *
 *  @pre    MLDmpOpen() \ifnot UMPL or MLDmpPedometerStandAloneOpen() \endif
 *          must have been called.
 *
 *  @param  data
 *              A pointer to an array to be passed back to the user.
 *              <b>Must be 3 cells long </b>.
 *
 *  @return Zero if the command is successful; an ML error code otherwise.
 */
tMLError MLGetAccelsFixed(long * data)
{
    tMLError result = ML_SUCCESS;
    if ( MLGetState() < ML_STATE_DMP_OPENED )
        return ML_ERROR_SM_IMPROPER_STATE;

    if (NULL == data) {
        return ML_ERROR_INVALID_PARAMETER;
    }
            result = FIFOGetAccel( data );

    return result;
}

/**
 *  @brief  MLGetTemperatureFixed is used to get temperature from FIFO.
 *
 *
 *  @pre    MLDmpOpen() \ifnot UMPL or MLDmpPedometerStandAloneOpen() \endif
 *          must have been called.
 *
 *  @param  data
 *              A pointer to the data to be passed back to the user.
 *
 *  @return Zero if the command is successful; an ML error code otherwise.
 */
tMLError MLGetTemperatureFixed(long * data)
{
    tMLError result = ML_SUCCESS;
    if ( MLGetState() < ML_STATE_DMP_OPENED )
        return ML_ERROR_SM_IMPROPER_STATE;

    if (NULL == data) {
        return ML_ERROR_INVALID_PARAMETER;
    }
            result = FIFOGetTemperature( data );

    return result;
}

/**
 *  @brief  MLGetRotationMatrixFixed is used to get Rotation Matrix from FIFO.
 *
 *
 *  @pre    MLDmpOpen() \ifnot UMPL or MLDmpPedometerStandAloneOpen() \endif
 *          must have been called.
 *
 *  @param  data
 *              A pointer to an array to be passed back to the user.
 *              <b>Must be 9 cells long</b>.
 *
 *  @return Zero if the command is successful; an ML error code otherwise.
 */
tMLError MLGetRotationMatrixFixed(long * data)
{
    tMLError result = ML_SUCCESS;
    long qdata[4];
    if ( MLGetState() < ML_STATE_DMP_OPENED )
        return ML_ERROR_SM_IMPROPER_STATE;

    if (NULL == data) {
        return ML_ERROR_INVALID_PARAMETER;
    }

            FIFOGetQuaternion(qdata);
            quaternionToRotationMatrix( qdata, data );

    return result;
}

/**
 *  @brief  MLGetQuaternionFixed is used to get Quaternion from FIFO.
 *
 *
 *  @pre    MLDmpOpen() \ifnot UMPL or MLDmpPedometerStandAloneOpen() \endif
 *          must have been called.
 *
 *  @param  data
 *              A pointer to an array to be passed back to the user.
 *              <b>Must be 4 cells long </b>.
 *
 *  @return Zero if the command is successful; an ML error code otherwise.
 */
tMLError MLGetQuaternionFixed(long * data)
{
    tMLError result = ML_SUCCESS;
    if ( MLGetState() < ML_STATE_DMP_OPENED )
        return ML_ERROR_SM_IMPROPER_STATE;

    if (NULL == data) {
        return ML_ERROR_INVALID_PARAMETER;
    }
            result = FIFOGetQuaternion(data);

    return result;
}

/**
 *  @brief  MLGetLinearAccelerationFixed is used to get accelerometer data from FIFO.
 *
 *
 *  @pre    MLDmpOpen() \ifnot UMPL or MLDmpPedometerStandAloneOpen() \endif
 *          must have been called.
 *
 *  @param  data
 *              A pointer to an array to be passed back to the user.
 *              <b>Must be 3 cells long </b>.
 *
 *  @return Zero if the command is successful; an ML error code otherwise.
 */
tMLError MLGetLinearAccelerationFixed(long * data)
{
    tMLError result = ML_SUCCESS;
    if ( MLGetState() < ML_STATE_DMP_OPENED )
        return ML_ERROR_SM_IMPROPER_STATE;

    if (NULL == data) {
        return ML_ERROR_INVALID_PARAMETER;
    }

            result = FIFOGetLinearAccel(data);

    return result;
}

/**
 *  @brief  MLGetLinearAccelerationWorldFixed is used to get accelerometer data from FIFO
 * in world frame with gravity removed.
 *
 *
 *  @pre    MLDmpOpen() \ifnot UMPL or MLDmpPedometerStandAloneOpen() \endif
 *          must have been called.
 *
 *  @param  data
 *              A pointer to an array to be passed back to the user.
 *              <b>Must be 3 cells long </b>.
 *
 *  @return Zero if the command is successful; an ML error code otherwise.
 */
tMLError MLGetLinearAccelerationWorldFixed(long * data)
{
    tMLError result = ML_SUCCESS;
    if ( MLGetState() < ML_STATE_DMP_OPENED )
        return ML_ERROR_SM_IMPROPER_STATE;

    if (NULL == data) {
        return ML_ERROR_INVALID_PARAMETER;
    }

        result = FIFOGetLinearAccelWorld(data);

    return result;
}

/**
 *  @brief  MLGetGravityFixed is used to get Gravity from FIFO
 *
 *
 *  @pre    MLDmpOpen() \ifnot UMPL or MLDmpPedometerStandAloneOpen() \endif
 *          must have been called.
 *
 *  @param  data
 *              A pointer to an array to be passed back to the user.
 *              <b>Must be 3 cells long</b>.
 *
 *  @return Zero if the command is successful; an ML error code otherwise.
 */
tMLError MLGetGravityFixed(long * data)
{
    tMLError result = ML_SUCCESS;
    if ( MLGetState() < ML_STATE_DMP_OPENED )
        return ML_ERROR_SM_IMPROPER_STATE;

    if (NULL == data) {
        return ML_ERROR_INVALID_PARAMETER;
    }

        result = FIFOGetGravBody( data );

    return result;
}

/**
 *  @brief  MLGetAngularVelocityFixed is used to get Angular Velocity
 *
 *
 *  @pre    MLDmpOpen() \ifnot UMPL or MLDmpPedometerStandAloneOpen() \endif
 *          must have been called.
 *
 *  @param  data
 *              A pointer to an array to be passed back to the user.
 *              <b>Must be 3 cells long </b>.
 *
 *  @return Zero if the command is successful; an ML error code otherwise.
 */
tMLError MLGetAngularVelocityFixed(long * data)
{
    tMLError result = ML_SUCCESS;
    if ( MLGetState() < ML_STATE_DMP_OPENED )
        return ML_ERROR_SM_IMPROPER_STATE;

    if (NULL == data) {
        return ML_ERROR_INVALID_PARAMETER;
    }

            data[0] = mlxData.mlAngVBody[0];
            data[1] = mlxData.mlAngVBody[1];
            data[2] = mlxData.mlAngVBody[2];

    return result;
}

/**
 *  @brief  MLGetEulerAnglesFixed is used to get Euler Angle from Rotation Matrix
 *
 *
 *  @pre    MLDmpOpen() \ifnot UMPL or MLDmpPedometerStandAloneOpen() \endif
 *          must have been called.
 *
 *  @param  data
 *              A pointer to an array to be passed back to the user.
 *              <b>Must be 3 cells long </b>.
 *
 *  @return Zero if the command is successful; an ML error code otherwise.
 */
tMLError MLGetEulerAnglesFixed(long * data)
{
    tMLError result = ML_SUCCESS;
    float rotMatrix[9];
	float tmp;
    if ( MLGetState() < ML_STATE_DMP_OPENED )
        return ML_ERROR_SM_IMPROPER_STATE;

    if (NULL == data) {
        return ML_ERROR_INVALID_PARAMETER;
    }

            result = MLGetRotationMatrixFloat(rotMatrix);
            tmp = rotMatrix[6];
            if (tmp>1.0f) {
                tmp = 1.0f;
            }
            if (tmp<-1.0f) {
                tmp = -1.0f;
            }
            data[0] = (long)((float)((double)atan2(rotMatrix[7], rotMatrix[8])*57.29577951308)*65536L);
            data[1] = (long)((float)((double)asin(tmp)*57.29577951308)*65536L);
            data[2] = (long)((float)((double)atan2(rotMatrix[3], rotMatrix[0])*57.29577951308)*65536L);


    return result;
}

/**
 *  @brief  MLGetEulerAnglesXFixed is used to get X Euler Angle from Rotation Matrix
 *
 *
 *  @pre    MLDmpOpen() \ifnot UMPL or MLDmpPedometerStandAloneOpen() \endif
 *          must have been called.
 *
 *  @param  data
 *              A pointer to an array to be passed back to the user.
 *              <b>Must be 3 cells long </b>.
 *
 *  @return Zero if the command is successful; an ML error code otherwise.
 */
tMLError MLGetEulerAnglesXFixed(long * data)
{
    tMLError result = ML_SUCCESS;
    float rotMatrix[9];
	float tmp;
    if ( MLGetState() < ML_STATE_DMP_OPENED )
        return ML_ERROR_SM_IMPROPER_STATE;

    if (NULL == data) {
        return ML_ERROR_INVALID_PARAMETER;
    }

            result = MLGetRotationMatrixFloat(rotMatrix);
            tmp = rotMatrix[6];
            if (tmp>1.0f) {
                tmp = 1.0f;
            }
            if (tmp<-1.0f) {
                tmp = -1.0f;
            }
            data[0] = (long)((float)((double)atan2(rotMatrix[7], rotMatrix[8])*57.29577951308)*65536L);
            data[1] = (long)((float)((double)asin(tmp)*57.29577951308)*65536L);
            data[2] = (long)((float)((double)atan2(rotMatrix[3], rotMatrix[0])*57.29577951308)*65536L);
    return result;
}

/**
 *  @brief  MLGetEulerAnglesYFixed is used to get Y Euler Angle from Rotation Matrix
 *
 *
 *  @pre    MLDmpOpen() \ifnot UMPL or MLDmpPedometerStandAloneOpen() \endif
 *          must have been called.
 *
 *  @param  data
 *              A pointer to an array to be passed back to the user.
 *              <b>Must be 3 cells long </b>.
 *
 *  @return Zero if the command is successful; an ML error code otherwise.
 */
tMLError MLGetEulerAnglesYFixed(long * data)
{
    tMLError result = ML_SUCCESS;
    float rotMatrix[9];
	float tmp;
    if ( MLGetState() < ML_STATE_DMP_OPENED )
        return ML_ERROR_SM_IMPROPER_STATE;

    if (NULL == data) {
        return ML_ERROR_INVALID_PARAMETER;
    }

            result = MLGetRotationMatrixFloat(rotMatrix);
            tmp = rotMatrix[7];
            if (tmp>1.0f) {
                tmp = 1.0f;
            }
            if (tmp<-1.0f) {
                tmp = -1.0f;
            }
            data[0] = (long)((float)((double)atan2(rotMatrix[8], rotMatrix[6])*57.29577951308f)*65536L);
            data[1] = (long)((float)((double)asin(tmp)*57.29577951308)*65536L);
            data[2] = (long)((float)((double)atan2(rotMatrix[4], rotMatrix[1])*57.29577951308f)*65536L);
    return result;
}

/**
 *  @brief  MLGetEulerAnglesZFixed is used to get Z Euler Angle from Rotation Matrix
 *
 *
 *  @pre    MLDmpOpen() \ifnot UMPL or MLDmpPedometerStandAloneOpen() \endif
 *          must have been called.
 *
 *  @param  data
 *              A pointer to an array to be passed back to the user.
 *              <b>Must be 3 cells long </b>.
 *
 *  @return Zero if the command is successful; an ML error code otherwise.
 */
tMLError MLGetEulerAnglesZFixed(long * data)
{
    tMLError result = ML_SUCCESS;
    float rotMatrix[9];
	float tmp;
    if ( MLGetState() < ML_STATE_DMP_OPENED )
        return ML_ERROR_SM_IMPROPER_STATE;

    if (NULL == data) {
        return ML_ERROR_INVALID_PARAMETER;
    }

            result = MLGetRotationMatrixFloat(rotMatrix);
            tmp = rotMatrix[8];
            if (tmp>1.0f) {
                tmp = 1.0f;
            }
            if (tmp<-1.0f) {
                tmp = -1.0f;
            }
            data[0] = (long)((float)((double)atan2(rotMatrix[6], rotMatrix[7])*57.29577951308)*65536L);
            data[1] = (long)((float)((double)asin(tmp)*57.29577951308)*65536L);
            data[2] = (long)((float)((double)atan2(rotMatrix[5], rotMatrix[2])*57.29577951308)*65536L);
    return result;
}

/**
 *  @brief  MLGetGyroTempSlopeFixed is used to get the temperature slope (dps per deg c)
 *
 *
 *  @pre    MLDmpOpen() \ifnot UMPL or MLDmpPedometerStandAloneOpen() \endif
 *          must have been called.
 *
 *  @param  data
 *              A pointer to an array to be passed back to the user.
 *              <b>Must be 3 cells long </b>.
 *
 *  @return Zero if the command is successful; an ML error code otherwise.
 */
tMLError MLGetGyroTempSlopeFixed(long * data)
{
    tMLError result = ML_SUCCESS;
    if ( MLGetState() < ML_STATE_DMP_OPENED )
        return ML_ERROR_SM_IMPROPER_STATE;

    if (NULL == data) {
        return ML_ERROR_INVALID_PARAMETER;
    }
           if (mlParams.biasUpdateFunc & ML_LEARN_BIAS_FROM_TEMPERATURE) {
                data[0] = (long)(mlxData.mlXGyroCoeff[1]*65536.0f);
                data[1] = (long)(mlxData.mlYGyroCoeff[1]*65536.0f);
                data[2] = (long)(mlxData.mlZGyroCoeff[1]*65536.0f);
            } else {
                data[0] = mlxData.mlTempSlope[0];
                data[1] = mlxData.mlTempSlope[1];
                data[2] = mlxData.mlTempSlope[2];
            }
    return result;
}

/**
 *  @brief  MLGetGyroBiasFixed is used to get Gyro Bias
 *
 *
 *  @pre    MLDmpOpen() \ifnot UMPL or MLDmpPedometerStandAloneOpen() \endif
 *          must have been called.
 *
 *  @param  data
 *              A pointer to an array to be passed back to the user.
 *              <b>Must be 3 cells long</b>.
 *
 *  @return Zero if the command is successful; an ML error code otherwise.
 */
tMLError MLGetGyroBiasFixed(long * data)
{
    tMLError result = ML_SUCCESS;
    if ( MLGetState() < ML_STATE_DMP_OPENED )
        return ML_ERROR_SM_IMPROPER_STATE;

    if (NULL == data) {
        return ML_ERROR_INVALID_PARAMETER;
    }
            data[0] = mlxData.mlGyroBias[0];
            data[1] = mlxData.mlGyroBias[1];
            data[2] = mlxData.mlGyroBias[2];

    return result;
}

/**
 *  @brief  MLGetAccelBiasFixed is used to get Accel Bias
 *
 *
 *  @pre    MLDmpOpen() \ifnot UMPL or MLDmpPedometerStandAloneOpen() \endif
 *          must have been called.
 *
 *  @param  data
 *              A pointer to an array to be passed back to the user.
 *              <b>Must be 3 cells long </b>.
 *
 *  @return Zero if the command is successful; an ML error code otherwise.
 */
tMLError MLGetAccelBiasFixed(long * data)
{
    tMLError result = ML_SUCCESS;
    if ( MLGetState() < ML_STATE_DMP_OPENED )
        return ML_ERROR_SM_IMPROPER_STATE;

    if (NULL == data) {
        return ML_ERROR_INVALID_PARAMETER;
    }
            data[0] = mlxData.mlGyroBias[3];
            data[1] = mlxData.mlGyroBias[4];
            data[2] = mlxData.mlGyroBias[5];

    return result;
}

/**
 *  @brief  MLGetMagBiasFixed is used to get Magnetometer Bias
 *
 *
 *  @pre    MLDmpOpen() \ifnot UMPL or MLDmpPedometerStandAloneOpen() \endif
 *          must have been called.
 *
 *  @param  data
 *              A pointer to an array to be passed back to the user.
 *              <b>Must be 3 cells long </b>.
 *
 *  @return Zero if the command is successful; an ML error code otherwise.
 */
tMLError MLGetMagBiasFixed(long * data)
{
    tMLError result = ML_SUCCESS;
    if ( MLGetState() < ML_STATE_DMP_OPENED )
        return ML_ERROR_SM_IMPROPER_STATE;

    if (NULL == data) {
        return ML_ERROR_INVALID_PARAMETER;
    }
            data[0] = mlxData.mlMagBias[0] + (long)((long long)mlxData.mlInitMagBias[0]* mlxData.mlMagSens / 16384);
            data[1] = mlxData.mlMagBias[1] + (long)((long long)mlxData.mlInitMagBias[1]* mlxData.mlMagSens / 16384);
            data[2] = mlxData.mlMagBias[2] + (long)((long long)mlxData.mlInitMagBias[2]* mlxData.mlMagSens / 16384);

    return result;
}

/**
 *  @brief  MLGetRawDataFixed is used to get Raw sensor data from FIFO.
 *
 *
 *  @pre    MLDmpOpen() \ifnot UMPL or MLDmpPedometerStandAloneOpen() \endif
 *          must have been called.
 *
 *  @param  data
 *              A pointer to an array to be passed back to the user.
 *              <b>Must be 9 cells long at least</b>.
 *
 *  @return Zero if the command is successful; an ML error code otherwise.
 */
tMLError MLGetRawDataFixed(long * data)
{
    tMLError result = ML_SUCCESS;
    if ( MLGetState() < ML_STATE_DMP_OPENED )
        return ML_ERROR_SM_IMPROPER_STATE;

    if (NULL == data) {
        return ML_ERROR_INVALID_PARAMETER;
    }

            result = FIFOGetSensorData( data );
            data[6] = mlxData.mlMagSensorData[0];
            data[7] = mlxData.mlMagSensorData[1];
            data[8] = mlxData.mlMagSensorData[2];

    return result;
}

/**
 *  @brief  MLGetMagRawDataFixed is used to get Raw magnetometer data.
 *
 *
 *  @pre    MLDmpOpen() \ifnot UMPL or MLDmpPedometerStandAloneOpen() \endif
 *          must have been called.
 *
 *  @param  data
 *              A pointer to an array to be passed back to the user.
 *              <b>Must be 3 cells long </b>.
 *
 *  @return Zero if the command is successful; an ML error code otherwise.
 */
tMLError MLGetMagRawDataFixed(long * data)
{
    tMLError result = ML_SUCCESS;
    if ( MLGetState() < ML_STATE_DMP_OPENED )
        return ML_ERROR_SM_IMPROPER_STATE;

    if (NULL == data) {
        return ML_ERROR_INVALID_PARAMETER;
    }

            data[0] = mlxData.mlMagSensorData[0];
            data[1] = mlxData.mlMagSensorData[1];
            data[2] = mlxData.mlMagSensorData[2];

    return result;
}

/**
 *  @brief  MLGetMagnetometerFixed is used to get magnetometer data.
 *
 *
 *  @pre    MLDmpOpen() \ifnot UMPL or MLDmpPedometerStandAloneOpen() \endif
 *          must have been called.
 *
 *  @param  data
 *              A pointer to an array to be passed back to the user.
 *              <b>Must be 3 cells long</b>.
 *
 *  @return Zero if the command is successful; an ML error code otherwise.
 */
tMLError MLGetMagnetometerFixed(long * data)
{
    tMLError result = ML_SUCCESS;
    if ( MLGetState() < ML_STATE_DMP_OPENED )
        return ML_ERROR_SM_IMPROPER_STATE;

    if (NULL == data) {
        return ML_ERROR_INVALID_PARAMETER;
    }

            data[0] = mlxData.mlMagSensorData[0] + mlxData.mlInitMagBias[0];
            data[1] = mlxData.mlMagSensorData[1] + mlxData.mlInitMagBias[1];
            data[2] = mlxData.mlMagSensorData[2] + mlxData.mlInitMagBias[2];

    return result;
}

/**
 *  @brief  MLGetPressureFixed is used to get Pressure data.
 *
 *
 *  @pre    MLDmpOpen() \ifnot UMPL or MLDmpPedometerStandAloneOpen() \endif
 *          must have been called.
 *
 *  @param  data
 *              A pointer to data to be passed back to the user.
 *
 *  @return Zero if the command is successful; an ML error code otherwise.
 */
tMLError MLGetPressureFixed(long * data)
{
    tMLError result = ML_SUCCESS;
    if ( MLGetState() < ML_STATE_DMP_OPENED )
        return ML_ERROR_SM_IMPROPER_STATE;

    if (NULL == data) {
        return ML_ERROR_INVALID_PARAMETER;
    }

            data[0] = mlxData.mlPressure;

    return result;
}

/**
 *  @brief  MLGetHeadingFixed is used to get heading from Rotation Matrix.
 *
 *
 *  @pre    MLDmpOpen() \ifnot UMPL or MLDmpPedometerStandAloneOpen() \endif
 *          must have been called.
 *
 *  @param  data
 *              A pointer to data to be passed back to the user.
 *
 *  @return Zero if the command is successful; an ML error code otherwise.
 */
tMLError MLGetHeadingFixed(long * data)
{
    tMLError result = ML_SUCCESS;
	float rotMatrix[9];
	float tmp;
    if ( MLGetState() < ML_STATE_DMP_OPENED )
        return ML_ERROR_SM_IMPROPER_STATE;

    if (NULL == data) {
        return ML_ERROR_INVALID_PARAMETER;
    }
            result = MLGetRotationMatrixFloat( rotMatrix );
            if ((rotMatrix[7]<0.707) && (rotMatrix[7]>-0.707)) {
                tmp = (float)((double)atan2(rotMatrix[4], rotMatrix[1])*57.29577951308 - 90.0f);
            } else {
                tmp = (float)((double)atan2(rotMatrix[5], rotMatrix[2])*57.29577951308 + 90.0f);
            }
            if (tmp<0) {
                tmp += 360.0f;
            }
            data[0] = (long)((360-tmp)*65536.0f);

    return result;
}

/**
 *  @brief  MLGetGyroCalibrationMatrixFixed is used to get gyro calibration matrix.
 *
 *
 *  @pre    MLDmpOpen() \ifnot UMPL or MLDmpPedometerStandAloneOpen() \endif
 *          must have been called.
 *
 *  @param  data
 *              A pointer to an array to be passed back to the user.
 *              <b>Must be 9 cells long at least</b>.
 *
 *  @return Zero if the command is successful; an ML error code otherwise.
 */
tMLError MLGetGyroCalibrationMatrixFixed(long * data)
{
    tMLError result = ML_SUCCESS;
    if ( MLGetState() < ML_STATE_DMP_OPENED )
        return ML_ERROR_SM_IMPROPER_STATE;

    if (NULL == data) {
        return ML_ERROR_INVALID_PARAMETER;
    }

            data[0] = mlxData.mlGyroCal[0];
            data[1] = mlxData.mlGyroCal[1];
            data[2] = mlxData.mlGyroCal[2];
            data[3] = mlxData.mlGyroCal[3];
            data[4] = mlxData.mlGyroCal[4];
            data[5] = mlxData.mlGyroCal[5];
            data[6] = mlxData.mlGyroCal[6];
            data[7] = mlxData.mlGyroCal[7];
            data[8] = mlxData.mlGyroCal[8];

    return result;
}

/**
 *  @brief  MLGetAccelCalibrationMatrixFixed is used to get accel calibration matrix.
 *
 *
 *  @pre    MLDmpOpen() \ifnot UMPL or MLDmpPedometerStandAloneOpen() \endif
 *          must have been called.
 *
 *  @param  data
 *              A pointer to an array to be passed back to the user.
 *              <b>Must be 9 cells long at least</b>.
 *
 *  @return Zero if the command is successful; an ML error code otherwise.
 */
tMLError MLGetAccelCalibrationMatrixFixed(long * data)
{
    tMLError result = ML_SUCCESS;
    if ( MLGetState() < ML_STATE_DMP_OPENED )
        return ML_ERROR_SM_IMPROPER_STATE;

    if (NULL == data) {
        return ML_ERROR_INVALID_PARAMETER;
    }

            data[0] = mlxData.mlAccelCal[0];
            data[1] = mlxData.mlAccelCal[1];
            data[2] = mlxData.mlAccelCal[2];
            data[3] = mlxData.mlAccelCal[3];
            data[4] = mlxData.mlAccelCal[4];
            data[5] = mlxData.mlAccelCal[5];
            data[6] = mlxData.mlAccelCal[6];
            data[7] = mlxData.mlAccelCal[7];
            data[8] = mlxData.mlAccelCal[8];

    return result;
}

/**
 *  @brief  MLGetMagCalibrationMatrixFixed is used to get magnetometer calibration matrix.
 *
 *
 *  @pre    MLDmpOpen() \ifnot UMPL or MLDmpPedometerStandAloneOpen() \endif
 *          must have been called.
 *
 *  @param  data
 *              A pointer to an array to be passed back to the user.
 *              <b>Must be 9 cells long at least</b>.
 *
 *  @return Zero if the command is successful; an ML error code otherwise.
 */
tMLError MLGetMagCalibrationMatrixFixed(long * data)
{
    tMLError result = ML_SUCCESS;
    if ( MLGetState() < ML_STATE_DMP_OPENED )
        return ML_ERROR_SM_IMPROPER_STATE;

    if (NULL == data) {
        return ML_ERROR_INVALID_PARAMETER;
    }

            data[0] = mlxData.mlMagCal[0];
            data[1] = mlxData.mlMagCal[1];
            data[2] = mlxData.mlMagCal[2];
            data[3] = mlxData.mlMagCal[3];
            data[4] = mlxData.mlMagCal[4];
            data[5] = mlxData.mlMagCal[5];
            data[6] = mlxData.mlMagCal[6];
            data[7] = mlxData.mlMagCal[7];
            data[8] = mlxData.mlMagCal[8];

    return result;
}

/**
 *  @brief  MLGetMagBiasErrorFixed is used to get magnetometer Bias error.
 *
 *
 *  @pre    MLDmpOpen() \ifnot UMPL or MLDmpPedometerStandAloneOpen() \endif
 *          must have been called.
 *
 *  @param  data
 *              A pointer to an array to be passed back to the user.
 *              <b>Must be 3 cells long at least</b>.
 *
 *  @return Zero if the command is successful; an ML error code otherwise.
 */
tMLError MLGetMagBiasErrorFixed(long * data)
{
    tMLError result = ML_SUCCESS;
    if ( MLGetState() < ML_STATE_DMP_OPENED )
        return ML_ERROR_SM_IMPROPER_STATE;

    if (NULL == data) {
        return ML_ERROR_INVALID_PARAMETER;
    }
            if (mlxData.mlLargeField==0) {
                data[0] = mlxData.mlMagBiasError[0];
                data[1] = mlxData.mlMagBiasError[1];
                data[2] = mlxData.mlMagBiasError[2];
            } else {
                data[0] = P_INIT;
                data[1] = P_INIT;
                data[2] = P_INIT;
            }

    return result;
}

/**
 *  @brief  MLGetMagScaleFixed is used to get magnetometer scale.
 *
 *
 *  @pre    MLDmpOpen() \ifnot UMPL or MLDmpPedometerStandAloneOpen() \endif
 *          must have been called.
 *
 *  @param  data
 *              A pointer to an array to be passed back to the user.
 *              <b>Must be 3 cells long at least</b>.
 *
 *  @return Zero if the command is successful; an ML error code otherwise.
 */
tMLError MLGetMagScaleFixed(long * data)
{
    tMLError result = ML_SUCCESS;
    if ( MLGetState() < ML_STATE_DMP_OPENED )
        return ML_ERROR_SM_IMPROPER_STATE;

    if (NULL == data) {
        return ML_ERROR_INVALID_PARAMETER;
    }

            data[0] = mlxData.mlMagScale[0];
            data[1] = mlxData.mlMagScale[1];
            data[2] = mlxData.mlMagScale[2];

    return result;
}

/**
 *  @brief  MLGetLocalFieldFixed is used to get local magnetic field data.
 *
 *
 *  @pre    MLDmpOpen() \ifnot UMPL or MLDmpPedometerStandAloneOpen() \endif
 *          must have been called.
 *
 *  @param  data
 *              A pointer to an array to be passed back to the user.
 *              <b>Must be 3 cells long at least</b>.
 *
 *  @return Zero if the command is successful; an ML error code otherwise.
 */
tMLError MLGetLocalFieldFixed(long * data)
{
    tMLError result = ML_SUCCESS;
    if ( MLGetState() < ML_STATE_DMP_OPENED )
        return ML_ERROR_SM_IMPROPER_STATE;

    if (NULL == data) {
        return ML_ERROR_INVALID_PARAMETER;
    }

            data[0] = mlxData.mlLocalField[0];
            data[1] = mlxData.mlLocalField[1];
            data[2] = mlxData.mlLocalField[2];

    return result;
}

/**
 *  @brief  MLGetRelativeQuaternionFixed is used to get relative quaternion.
 *
 *
 *  @pre    MLDmpOpen() \ifnot UMPL or MLDmpPedometerStandAloneOpen() \endif
 *          must have been called.
 *
 *  @param  data
 *              A pointer to an array to be passed back to the user.
 *              <b>Must be 4 cells long at least</b>.
 *
 *  @return Zero if the command is successful; an ML error code otherwise.
 */
tMLError MLGetRelativeQuaternionFixed(long * data)
{
    tMLError result = ML_SUCCESS;
    if ( MLGetState() < ML_STATE_DMP_OPENED )
        return ML_ERROR_SM_IMPROPER_STATE;

    if (NULL == data) {
        return ML_ERROR_INVALID_PARAMETER;
    }

            data[0] = mlxData.mlRelativeQuat[0];
            data[1] = mlxData.mlRelativeQuat[1];
            data[2] = mlxData.mlRelativeQuat[2];
            data[3] = mlxData.mlRelativeQuat[3];

    return result;
}

/**
 *  @brief  MLGetGyrosFloat is used to get an array of three data points representing the
 *			X gyroscope, Y gyroscope, and Z gyroscope values.
 *          The values are not sensor fused with other sensor types data but
 *          reflect the orientation from the mounting matrices in use.
 *          The ML_GYROS values are scaled to ensure 1 dps corresponds to 2^16
 *          codes.
 *
 *  @pre    MLDmpOpen() \ifnot UMPL or MLDmpPedometerStandAloneOpen() \endif
 *          must have been called.
 *
 *  @param  data
 *              A pointer to an array to be passed back to the user.
 *              <b>Must be 3 cells long</b>.
 *
 *  @return ML_SUCCESS if the command is successful; an error code otherwise.
 */
tMLError MLGetGyrosFloat(float *data)
{
    INVENSENSE_FUNC_START;

    tMLError result = ML_SUCCESS;
    long ldata[3];

    if ( MLGetState() < ML_STATE_DMP_OPENED )
        return ML_ERROR_SM_IMPROPER_STATE;

    if (NULL == data) {
        return ML_ERROR_INVALID_PARAMETER;
    }

    result = FIFOGetGyro(ldata);
    data[0] = (float)ldata[0]/65536.0f;
    data[1] = (float)ldata[1]/65536.0f;
    data[2] = (float)ldata[2]/65536.0f;

    return result;
}

/**
 *  @brief  MLGetAngularVelocityFloat is used to get an array of three data points representing the angular
 *          velocity as derived from <b>both</b> gyroscopes and accelerometers.
 *          This requires that ML_SENSOR_FUSION be enabled, to fuse data from
 *          the gyroscope and accelerometer device, appropriately scaled and
 *          oriented according to the respective mounting matrices.
 *
 *  @pre    MLDmpOpen() \ifnot UMPL or MLDmpPedometerStandAloneOpen() \endif
 *          must have been called.
 *  @param  data
 *              A pointer to an array to be passed back to the user.
 *              <b>Must be 3 cells long</b>.
 *
 *  @return ML_SUCCESS if the command is successful; an error code otherwise.
 */
tMLError MLGetAngularVelocityFloat(float *data)
{
    return MLGetGyrosFloat(data);
}

/**
 *  @brief  MLGetAccelsFloat is used to get an array of three data points representing the X
 *          accelerometer, Y accelerometer, and Z accelerometer values.
 *          The values are not sensor fused with other sensor types data but
 *          reflect the orientation from the mounting matrices in use.
 *          The ML_ACCELS values are scaled to ensure 1 gee corresponds to 2^16
 *          codes.
 *
 *  @pre    MLDmpOpen() \ifnot UMPL or MLDmpPedometerStandAloneOpen() \endif
 *          must have been called.
 *
 *  @param  data
 *              A pointer to an array to be passed back to the user.
 *              <b>Must be 3 cells long at least</b>.
 *
 *  @return ML_SUCCESS if the command is successful; an error code otherwise.
 */
tMLError MLGetAccelsFloat(float *data)
{
    INVENSENSE_FUNC_START;

    tMLError result = ML_SUCCESS;

    if ( MLGetState() < ML_STATE_DMP_OPENED )
        return ML_ERROR_SM_IMPROPER_STATE;

    if (NULL == data) {
        return ML_ERROR_INVALID_PARAMETER;
    }

    result = FIFOGetAccelFloat( data );

    return result;
}

/**
 *  @brief  MLGetTemperatureFloat is used to get temperature from FIFO.
 *
 *  @pre    MLDmpOpen() \ifnot UMPL or MLDmpPedometerStandAloneOpen() \endif
 *          must have been called.
 *
 *  @param  data
 *              A pointer to data to be passed back to the user.
 *
 *  @return Zero if the command is successful; an ML error code otherwise.
 */
tMLError MLGetTemperatureFloat(float *data)
{
    INVENSENSE_FUNC_START;
    tMLError result = ML_SUCCESS;
    long ldata[1];

    if (MLGetState() < ML_STATE_DMP_OPENED)
        return ML_ERROR_SM_IMPROPER_STATE;

    if (NULL == data) {
        return ML_ERROR_INVALID_PARAMETER;
    }

    result = FIFOGetTemperature(ldata);
    data[0] = (float)ldata[0] / 65536.0f;

    return result;
}

/**
 *  @brief  MLGetRotationMatrixFloat is used to get an array of nine data points representing the rotation
 *          matrix generated from all available sensors.
 *          This requires that ML_SENSOR_FUSION be enabled.
 *          The array format will be R11, R12, R13, R21, R22, R23, R31, R32,
 *          R33, representing the matrix:
 *          <center>R11 R12 R13</center>
 *          <center>R21 R22 R23</center>
 *          <center>R31 R32 R33</center>
 *          <b>Please refer to the "9-Axis Sensor Fusion Application Note" document,
 *          section 7 "Sensor Fusion Output", for details regarding rotation
 *          matrix output</b>.
 *
 *  @pre    MLDmpOpen() \ifnot UMPL or MLDmpPedometerStandAloneOpen() \endif
 *          must have been called.
 *
 *  @param  data
 *              A pointer to an array to be passed back to the user.
 *              <b>Must be 9 cells long at least</b>.
 *
 *  @return ML_SUCCESS if the command is successful; an error code otherwise.
 */
tMLError MLGetRotationMatrixFloat(float *data)
{
    INVENSENSE_FUNC_START;

    tMLError result = ML_SUCCESS;

    if ( MLGetState() < ML_STATE_DMP_OPENED )
        return ML_ERROR_SM_IMPROPER_STATE;

    if (NULL == data) {
        return ML_ERROR_INVALID_PARAMETER;
    }
            {
                long qdata[4],rdata[9];
                FIFOGetQuaternion(qdata);
                quaternionToRotationMatrix( qdata, rdata );
                data[0] = (float)rdata[0]/1073741824.0f;
                data[1] = (float)rdata[1]/1073741824.0f;
                data[2] = (float)rdata[2]/1073741824.0f;
                data[3] = (float)rdata[3]/1073741824.0f;
                data[4] = (float)rdata[4]/1073741824.0f;
                data[5] = (float)rdata[5]/1073741824.0f;
                data[6] = (float)rdata[6]/1073741824.0f;
                data[7] = (float)rdata[7]/1073741824.0f;
                data[8] = (float)rdata[8]/1073741824.0f;
            }

    return result;
}

/**
 *  @brief  MLGetQuaternionFloat is used to get an array of four data points representing the quaternion
 *          generated from all available sensors.
 *          This requires that ML_SENSOR_FUSION be enabled.
 *
 *  @pre    MLDmpOpen() \ifnot UMPL or MLDmpPedometerStandAloneOpen() \endif
 *          must have been called.
 *
 *  @param  data
 *              A pointer to an array to be passed back to the user.
 *              <b>Must be 4 cells long at least</b>.
 *
 *  @return ML_SUCCESS if the command is successful; an error code otherwise.
 */
tMLError MLGetQuaternionFloat(float *data)
{
    INVENSENSE_FUNC_START;

    tMLError result = ML_SUCCESS;
	long ldata[4];

    if ( MLGetState() < ML_STATE_DMP_OPENED )
        return ML_ERROR_SM_IMPROPER_STATE;

    if (NULL == data) {
        return ML_ERROR_INVALID_PARAMETER;
    }
			result = FIFOGetQuaternion(ldata);
            data[0] = (float)ldata[0]/1073741824.0f;
            data[1] = (float)ldata[1]/1073741824.0f;
            data[2] = (float)ldata[2]/1073741824.0f;
            data[3] = (float)ldata[3]/1073741824.0f;

    return result;
}

/**
 *  @brief  MLGetLinearAccelerationFloat is used to get an array of three data points representing
 *			the linear acceleration as derived from both gyroscopes and accelerometers.
 *          This requires that ML_SENSOR_FUSION be enabled.
 *
 *  @pre    MLDmpOpen() \ifnot UMPL or MLDmpPedometerStandAloneOpen() \endif
 *          must have been called.
 *
 *  @param  data
 *              A pointer to an array to be passed back to the user.
 *              <b>Must be 3 cells long at least</b>.
 *
 *  @return ML_SUCCESS if the command is successful; an error code otherwise.
 */
tMLError MLGetLinearAccelerationFloat(float *data)
{
    INVENSENSE_FUNC_START;

    tMLError result = ML_SUCCESS;
	long ldata[3];

    if ( MLGetState() < ML_STATE_DMP_OPENED )
        return ML_ERROR_SM_IMPROPER_STATE;

    if (NULL == data) {
        return ML_ERROR_INVALID_PARAMETER;
    }

            result = FIFOGetLinearAccel( ldata );
            data[0] = (float)ldata[0]/65536.0f;
            data[1] = (float)ldata[1]/65536.0f;
            data[2] = (float)ldata[2]/65536.0f;

    return result;
}

/**
 *  @brief  MLGetLinearAccelerationWorldFloat is used to get an array of three data points
 *			representing linear acceleration in world coordinates, as derived from both gyroscopes
 *          and accelerometers.
 *          This requires that ML_SENSOR_FUSION be enabled.
 *
 *  @pre    MLDmpOpen() \ifnot UMPL or MLDmpPedometerStandAloneOpen() \endif
 *          must have been called.
 *
 *  @param  data
 *              A pointer to an array to be passed back to the user.
 *              <b>Must be 3 cells long at least</b>.
 *
 *  @return ML_SUCCESS if the command is successful; an error code otherwise.
 */
tMLError MLGetLinearAccelerationWorldFloat(float *data)
{
    INVENSENSE_FUNC_START;

    tMLError result = ML_SUCCESS;
	long ldata[3];

    if ( MLGetState() < ML_STATE_DMP_OPENED )
        return ML_ERROR_SM_IMPROPER_STATE;

    if (NULL == data) {
        return ML_ERROR_INVALID_PARAMETER;
    }

            result = FIFOGetLinearAccelWorld( ldata );
            data[0] = (float)ldata[0]/65536.0f;
            data[1] = (float)ldata[1]/65536.0f;
            data[2] = (float)ldata[2]/65536.0f;

    return result;
}

/**
 *  @brief  MLGetGravityFloat is used to get an array of three data points
 *			representing the direction of gravity in body coordinates, as derived from
 *			both gyroscopes and accelerometers.
 *          This requires that ML_SENSOR_FUSION be enabled.
 *  @pre    MLDmpOpen() \ifnot UMPL or MLDmpPedometerStandAloneOpen() \endif
 *          must have been called.
 *
 *  @param  data
 *              A pointer to an array to be passed back to the user.
 *              <b>Must be 3 cells long at least</b>.
 *
 *  @return ML_SUCCESS if the command is successful; an error code otherwise.
 */
tMLError MLGetGravityFloat(float *data)
{
    INVENSENSE_FUNC_START;

    tMLError result = ML_SUCCESS;
	long ldata[3];

    if ( MLGetState() < ML_STATE_DMP_OPENED )
        return ML_ERROR_SM_IMPROPER_STATE;

    if (NULL == data) {
        return ML_ERROR_INVALID_PARAMETER;
    }
            result = FIFOGetGravBody( ldata );
            data[0] = (float)ldata[0]/65536.0f;
            data[1] = (float)ldata[1]/65536.0f;
            data[2] = (float)ldata[2]/65536.0f;

    return result;
}

/**
 *  @brief  MLGetGyroCalibrationMatrixFloat is used to get an array of nine data points
 *			representing the calibration matrix for the gyroscopes:
 *          <center>C11 C12 C13</center>
 *          <center>C21 C22 C23</center>
 *          <center>C31 C32 C33</center>
 *  @pre    MLDmpOpen() \ifnot UMPL or MLDmpPedometerStandAloneOpen() \endif
 *          must have been called.
 *
 *  @param  data
 *              A pointer to an array to be passed back to the user.
 *              <b>Must be 9 cells long at least</b>.
 *
 *  @return ML_SUCCESS if the command is successful; an error code otherwise.
 */
tMLError MLGetGyroCalibrationMatrixFloat(float *data)
{
    INVENSENSE_FUNC_START;

    tMLError result = ML_SUCCESS;

    if ( MLGetState() < ML_STATE_DMP_OPENED )
        return ML_ERROR_SM_IMPROPER_STATE;

    if (NULL == data) {
        return ML_ERROR_INVALID_PARAMETER;
    }

            data[0] = (float)mlxData.mlGyroCal[0]/1073741824.0f;
            data[1] = (float)mlxData.mlGyroCal[1]/1073741824.0f;
            data[2] = (float)mlxData.mlGyroCal[2]/1073741824.0f;
            data[3] = (float)mlxData.mlGyroCal[3]/1073741824.0f;
            data[4] = (float)mlxData.mlGyroCal[4]/1073741824.0f;
            data[5] = (float)mlxData.mlGyroCal[5]/1073741824.0f;
            data[6] = (float)mlxData.mlGyroCal[6]/1073741824.0f;
            data[7] = (float)mlxData.mlGyroCal[7]/1073741824.0f;
            data[8] = (float)mlxData.mlGyroCal[8]/1073741824.0f;

    return result;
}

/**
 *  @brief  MLGetAccelCalibrationMatrixFloat is used to get an array of nine data points
 *			representing the calibration matrix for the accelerometers:
 *          <center>C11 C12 C13</center>
 *          <center>C21 C22 C23</center>
 *          <center>C31 C32 C33</center>
 *
 *  @pre    MLDmpOpen() \ifnot UMPL or MLDmpPedometerStandAloneOpen() \endif
 *          must have been called.
 *
 *  @param  data
 *              A pointer to an array to be passed back to the user.
 *              <b>Must be 9 cells long at least</b>.
 *
 *  @return ML_SUCCESS if the command is successful; an error code otherwise.
 */
tMLError MLGetAccelCalibrationMatrixFloat(float *data)
{
    INVENSENSE_FUNC_START;

    tMLError result = ML_SUCCESS;

    if ( MLGetState() < ML_STATE_DMP_OPENED )
        return ML_ERROR_SM_IMPROPER_STATE;

    if (NULL == data) {
        return ML_ERROR_INVALID_PARAMETER;
    }

            data[0] = (float)mlxData.mlAccelCal[0]/1073741824.0f;
            data[1] = (float)mlxData.mlAccelCal[1]/1073741824.0f;
            data[2] = (float)mlxData.mlAccelCal[2]/1073741824.0f;
            data[3] = (float)mlxData.mlAccelCal[3]/1073741824.0f;
            data[4] = (float)mlxData.mlAccelCal[4]/1073741824.0f;
            data[5] = (float)mlxData.mlAccelCal[5]/1073741824.0f;
            data[6] = (float)mlxData.mlAccelCal[6]/1073741824.0f;
            data[7] = (float)mlxData.mlAccelCal[7]/1073741824.0f;
            data[8] = (float)mlxData.mlAccelCal[8]/1073741824.0f;

    return result;
}

/**
 *  @brief  MLGetMagCalibrationMatrixFloat is used to get an array of nine data points
 *			representing the calibration matrix for the compass:
 *          <center>C11 C12 C13</center>
 *          <center>C21 C22 C23</center>
 *          <center>C31 C32 C33</center>
 *
 *  @pre    MLDmpOpen() \ifnot UMPL or MLDmpPedometerStandAloneOpen() \endif
 *          must have been called.
 *
 *  @param  data
 *              A pointer to an array to be passed back to the user.
 *              <b>Must be 9 cells long at least</b>.
 *
 *  @return ML_SUCCESS if the command is successful; an error code otherwise.
 */
tMLError MLGetMagCalibrationMatrixFloat(float *data)
{
    INVENSENSE_FUNC_START;

    tMLError result = ML_SUCCESS;

    if ( MLGetState() < ML_STATE_DMP_OPENED )
        return ML_ERROR_SM_IMPROPER_STATE;

    if (NULL == data) {
        return ML_ERROR_INVALID_PARAMETER;
    }

            data[0] = (float)mlxData.mlMagCal[0]/1073741824.0f;
            data[1] = (float)mlxData.mlMagCal[1]/1073741824.0f;
            data[2] = (float)mlxData.mlMagCal[2]/1073741824.0f;
            data[3] = (float)mlxData.mlMagCal[3]/1073741824.0f;
            data[4] = (float)mlxData.mlMagCal[4]/1073741824.0f;
            data[5] = (float)mlxData.mlMagCal[5]/1073741824.0f;
            data[6] = (float)mlxData.mlMagCal[6]/1073741824.0f;
            data[7] = (float)mlxData.mlMagCal[7]/1073741824.0f;
            data[8] = (float)mlxData.mlMagCal[8]/1073741824.0f;
    return result;
}

/**
 *  @brief  MLGetGyroTempSlopeFloat is used to get is used to get the temperature slope
 *  @pre    MLDmpOpen() \ifnot UMPL or MLDmpPedometerStandAloneOpen() \endif
 *          must have been called.
 *
 *  @param  data
 *              A pointer to an array to be passed back to the user.
 *              <b>Must be 3 cells long </b>.
 *
 *  @return ML_SUCCESS if the command is successful; an error code otherwise.
 */
tMLError MLGetGyroTempSlopeFloat(float *data)
{
    INVENSENSE_FUNC_START;

    tMLError result = ML_SUCCESS;

    if ( MLGetState() < ML_STATE_DMP_OPENED )
        return ML_ERROR_SM_IMPROPER_STATE;

    if (NULL == data) {
        return ML_ERROR_INVALID_PARAMETER;
    }

            if (mlParams.biasUpdateFunc & ML_LEARN_BIAS_FROM_TEMPERATURE) {
                data[0] = mlxData.mlXGyroCoeff[1];
                data[1] = mlxData.mlYGyroCoeff[1];
                data[2] = mlxData.mlZGyroCoeff[1];
            } else {
                data[0] = (float)mlxData.mlTempSlope[0]/65536.0f;
                data[1] = (float)mlxData.mlTempSlope[1]/65536.0f;
                data[2] = (float)mlxData.mlTempSlope[2]/65536.0f;
            }


    return result;
}

/**
 *  @brief  MLGetGyroBiasFloat is used to get an array of three data points representing
 *			the gyroscope biases.
 *  @pre    MLDmpOpen() \ifnot UMPL or MLDmpPedometerStandAloneOpen() \endif
 *          must have been called.
 *
 *  @param  data
 *              A pointer to an array to be passed back to the user.
 *              <b>Must be 3 cells long</b>.
 *
 *  @return ML_SUCCESS if the command is successful; an error code otherwise.
 */
tMLError MLGetGyroBiasFloat(float *data)
{
    INVENSENSE_FUNC_START;

    tMLError result = ML_SUCCESS;

    if ( MLGetState() < ML_STATE_DMP_OPENED )
        return ML_ERROR_SM_IMPROPER_STATE;

    if (NULL == data) {
        return ML_ERROR_INVALID_PARAMETER;
    }

            data[0] = (float)mlxData.mlGyroBias[0]/65536.0f;
            data[1] = (float)mlxData.mlGyroBias[1]/65536.0f;
            data[2] = (float)mlxData.mlGyroBias[2]/65536.0f;

    return result;
}

/**
 *  @brief  MLGetAccelBiasFloat is used to get an array of three data points representing
 *			the accelerometer biases.
 *
 *  @pre    MLDmpOpen() \ifnot UMPL or MLDmpPedometerStandAloneOpen() \endif
 *          must have been called.
 *
 *  @param  data
 *              A pointer to an array to be passed back to the user.
 *              <b>Must be 3 cells long</b>.
 *
 *  @return ML_SUCCESS if the command is successful; an error code otherwise.
 */
tMLError MLGetAccelBiasFloat(float *data)
{
    INVENSENSE_FUNC_START;

    tMLError result = ML_SUCCESS;

    if ( MLGetState() < ML_STATE_DMP_OPENED )
        return ML_ERROR_SM_IMPROPER_STATE;

    if (NULL == data) {
        return ML_ERROR_INVALID_PARAMETER;
    }

            data[0] = (float)mlxData.mlGyroBias[3]/65536.0f;
            data[1] = (float)mlxData.mlGyroBias[4]/65536.0f;
            data[2] = (float)mlxData.mlGyroBias[5]/65536.0f;

    return result;
}

/**
 *  @brief  MLGetAccelBiasFloat is used to get an array of three data points representing
 *			the compass biases.
 *  @pre    MLDmpOpen() \ifnot UMPL or MLDmpPedometerStandAloneOpen() \endif
 *          must have been called.
 *
 *  @param  data
 *              A pointer to an array to be passed back to the user.
 *              <b>Must be 3 cells long</b>.
 *
 *  @return ML_SUCCESS if the command is successful; an error code otherwise.
 */
tMLError MLGetMagBiasFloat(float *data)
{
    INVENSENSE_FUNC_START;

    tMLError result = ML_SUCCESS;

    if ( MLGetState() < ML_STATE_DMP_OPENED )
        return ML_ERROR_SM_IMPROPER_STATE;

    if (NULL == data) {
        return ML_ERROR_INVALID_PARAMETER;
    }

            data[0] = ((float)(mlxData.mlMagBias[0] + (long)((long long)mlxData.mlInitMagBias[0]* mlxData.mlMagSens / 16384)))/65536.0f;
            data[1] = ((float)(mlxData.mlMagBias[1] + (long)((long long)mlxData.mlInitMagBias[1]* mlxData.mlMagSens / 16384)))/65536.0f;
            data[2] = ((float)(mlxData.mlMagBias[2] + (long)((long long)mlxData.mlInitMagBias[2]* mlxData.mlMagSens / 16384)))/65536.0f;

    return result;
}

/**
 *  @brief  MLGetRawDataFloat is used to get an array of nine data points representing raw sensor data
 *          of the gyroscope X, Y, Z, accelerometer X, Y, Z, and
 *          compass X, Y, Z values.
 *          These values are not scaled and come out directly from the devices'
 *          sensor data output. In case of accelerometers with lower output
 *          resolution, e.g 8-bit, the sensor data is scaled up to match the
 *          2^14 = 1 gee typical representation for a +/- 2 gee full scale
 *          range.
 *
 *  @pre    MLDmpOpen() \ifnot UMPL or MLDmpPedometerStandAloneOpen() \endif
 *          must have been called.
 *
 *  @param  data
 *              A pointer to an array to be passed back to the user.
 *              <b>Must be 9 cells long at least</b>.
 *
 *  @return ML_SUCCESS if the command is successful; an error code otherwise.
 */
tMLError MLGetRawDataFloat(float *data)
{
    INVENSENSE_FUNC_START;

    tMLError result = ML_SUCCESS;
	long ldata[6];

    if ( MLGetState() < ML_STATE_DMP_OPENED )
        return ML_ERROR_SM_IMPROPER_STATE;

    if (NULL == data) {
        return ML_ERROR_INVALID_PARAMETER;
    }

            result = FIFOGetSensorData( ldata );
            data[0] = (float)ldata[0];
            data[1] = (float)ldata[1];
            data[2] = (float)ldata[2];
            data[3] = (float)ldata[3];
            data[4] = (float)ldata[4];
            data[5] = (float)ldata[5];
            data[6] = (float)mlxData.mlMagSensorData[0];
            data[7] = (float)mlxData.mlMagSensorData[1];
            data[8] = (float)mlxData.mlMagSensorData[2];

    return result;
}


/**
 *  @brief  MLGetEulerAnglesXFloat is used to get an array of three data points representing roll, pitch, and
 *          yaw using the X axis of the gyroscope, accelerometer, and compass as
 *          reference axis.
 *          This is typically the convention used for mobile devices where the X
 *          axis is the width of the screen, Y axis is the height, and Z the
 *          depth. In this case roll is defined as the rotation around the X
 *          axis of the device.
 *          The euler angles convention for this output is the following:
 *          <TABLE>
 *          <TR><TD><b>EULER ANGLE</b></TD><TD><b>ROTATION AROUND</b></TD></TR>
 *          <TR><TD>roll              </TD><TD>X axis                </TD></TR>
 *          <TR><TD>pitch             </TD><TD>Y axis                </TD></TR>
 *          <TR><TD>yaw               </TD><TD>Z axis                </TD></TR>
 *          </TABLE>
 *          ML_EULER_ANGLES_X corresponds to the ML_EULER_ANGLES output and is
 *          therefore the default convention.
 *
 *  @pre    MLDmpOpen() \ifnot UMPL or MLDmpPedometerStandAloneOpen() \endif
 *          must have been called.
 *
 *  @param  data
 *              A pointer to an array to be passed back to the user.
 *              <b>Must be 3 cells long</b>.
 *
 *  @return ML_SUCCESS if the command is successful; an error code otherwise.
 */
tMLError MLGetEulerAnglesXFloat(float *data)
{
    INVENSENSE_FUNC_START;

    tMLError result = ML_SUCCESS;
    float rotMatrix[9];
	float tmp;

    if ( MLGetState() < ML_STATE_DMP_OPENED )
        return ML_ERROR_SM_IMPROPER_STATE;

    if (NULL == data) {
        return ML_ERROR_INVALID_PARAMETER;
    }

            result = MLGetRotationMatrixFloat( rotMatrix );
            tmp = rotMatrix[6];
            if (tmp>1.0f) {
                tmp = 1.0f;
            }
            if (tmp<-1.0f) {
                tmp = -1.0f;
            }
            data[0] = (float)((double)atan2(rotMatrix[7], rotMatrix[8])*57.29577951308);
            data[1] = (float)((double)asin(tmp)*57.29577951308);
            data[2] = (float)((double)atan2(rotMatrix[3], rotMatrix[0])*57.29577951308);

    return result;
}

/**
 *  @brief  MLGetEulerAnglesFloat is used to get an array of three data points three data points
 *			representing roll, pitch, and yaw corresponding to the ML_EULER_ANGLES_X output and it is
 *          therefore the default convention for Euler angles.
 *          Please refer to the ML_EULER_ANGLES_X for a detailed description.
 *
 *  @pre    MLDmpOpen() \ifnot UMPL or MLDmpPedometerStandAloneOpen() \endif
 *          must have been called.
 *
 *  @param  data
 *              A pointer to an array to be passed back to the user.
 *              <b>Must be 3 cells long</b>.
 *
 *  @return ML_SUCCESS if the command is successful; an error code otherwise.
 */
tMLError MLGetEulerAnglesFloat(float *data)
{
    return MLGetEulerAnglesXFloat(data);
}

/**
 *  @brief  MLGetEulerAnglesYFloat is used to get an array of three data points representing roll, pitch, and
 *          yaw using the Y axis of the gyroscope, accelerometer, and compass as
 *          reference axis.
 *          This convention is typically used in augmented reality applications,
 *          where roll is defined as the rotation around the axis along the
 *          height of the screen of a mobile device, namely the Y axis.
 *          The euler angles convention for this output is the following:
 *          <TABLE>
 *          <TR><TD><b>EULER ANGLE</b></TD><TD><b>ROTATION AROUND</b></TD></TR>
 *          <TR><TD>roll              </TD><TD>Y axis                </TD></TR>
 *          <TR><TD>pitch             </TD><TD>X axis                </TD></TR>
 *          <TR><TD>yaw               </TD><TD>Z axis                </TD></TR>
 *          </TABLE>
 *
 *  @pre    MLDmpOpen() \ifnot UMPL or MLDmpPedometerStandAloneOpen() \endif
 *          must have been called.
 *
 *  @param  data
 *              A pointer to an array to be passed back to the user.
 *              <b>Must be 3 cells long</b>.
 *
 *  @return ML_SUCCESS if the command is successful; an error code otherwise.
 */
tMLError MLGetEulerAnglesYFloat(float *data)
{
    INVENSENSE_FUNC_START;

    tMLError result = ML_SUCCESS;
    float rotMatrix[9];
	float tmp;

    if ( MLGetState() < ML_STATE_DMP_OPENED )
        return ML_ERROR_SM_IMPROPER_STATE;

    if (NULL == data) {
        return ML_ERROR_INVALID_PARAMETER;
    }

            result = MLGetRotationMatrixFloat(rotMatrix);
            tmp = rotMatrix[7];
            if (tmp>1.0f) {
                tmp = 1.0f;
            }
            if (tmp<-1.0f) {
                tmp = -1.0f;
            }
            data[0] = (float)((double)atan2(rotMatrix[8], rotMatrix[6])*57.29577951308);
            data[1] = (float)((double)asin(tmp)*57.29577951308);
            data[2] = (float)((double)atan2(rotMatrix[4], rotMatrix[1])*57.29577951308);

    return result;
}

/**
 *  @brief  MLGetEulerAnglesZFloat is used to get an array of three data points representing roll, pitch, and
 *          yaw using the Z axis of the gyroscope, accelerometer, and compass as
 *          reference axis.
 *          This convention is mostly used in application involving the use
 *          of a camera, typically placed on the back of a mobile device, that
 *          is along the Z axis.  In this convention roll is defined as the
 *          rotation around the Z axis.
 *          The euler angles convention for this output is the following:
 *          <TABLE>
 *          <TR><TD><b>EULER ANGLE</b></TD><TD><b>ROTATION AROUND</b></TD></TR>
 *          <TR><TD>roll              </TD><TD>Z axis                </TD></TR>
 *          <TR><TD>pitch             </TD><TD>X axis                </TD></TR>
 *          <TR><TD>yaw               </TD><TD>Y axis                </TD></TR>
 *          </TABLE>
 *
 *  @pre    MLDmpOpen() \ifnot UMPL or MLDmpPedometerStandAloneOpen() \endif
 *          must have been called.
 *
 *  @param  data
 *              A pointer to an array to be passed back to the user.
 *              <b>Must be 3 cells long</b>.
 *
 *  @return ML_SUCCESS if the command is successful; an error code otherwise.
 */
tMLError MLGetEulerAnglesZFloat(float *data)
{
    INVENSENSE_FUNC_START;

    tMLError result = ML_SUCCESS;
    float rotMatrix[9];
	float tmp;

    if ( MLGetState() < ML_STATE_DMP_OPENED )
        return ML_ERROR_SM_IMPROPER_STATE;

    if (NULL == data) {
        return ML_ERROR_INVALID_PARAMETER;
    }

            result = MLGetRotationMatrixFloat(rotMatrix);
            tmp = rotMatrix[8];
            if (tmp>1.0f) {
                tmp = 1.0f;
            }
            if (tmp<-1.0f) {
                tmp = -1.0f;
            }
            data[0] = (float)((double)atan2(rotMatrix[6], rotMatrix[7])*57.29577951308);
            data[1] = (float)((double)asin(tmp)*57.29577951308);
            data[2] = (float)((double)atan2(rotMatrix[5], rotMatrix[2])*57.29577951308);

    return result;
}

/**
 *  @brief  MLGetMagRawDataFloat is used to get Raw magnetometer data
 *
 *  @pre    MLDmpOpen() \ifnot UMPL or MLDmpPedometerStandAloneOpen() \endif
 *          must have been called.
 *
 *  @param  data
 *              A pointer to an array to be passed back to the user.
 *              <b>Must be 3 cells long</b>.
 *
 *  @return ML_SUCCESS if the command is successful; an error code otherwise.
 */
tMLError MLGetMagRawDataFloat(float *data)
{
    INVENSENSE_FUNC_START;

    tMLError result = ML_SUCCESS;

    if ( MLGetState() < ML_STATE_DMP_OPENED )
        return ML_ERROR_SM_IMPROPER_STATE;

    if (NULL == data) {
        return ML_ERROR_INVALID_PARAMETER;
    }

            data[0] = (float)(mlxData.mlMagSensorData[0] + mlxData.mlInitMagBias[0]);
            data[1] = (float)(mlxData.mlMagSensorData[1] + mlxData.mlInitMagBias[1]);
            data[2] = (float)(mlxData.mlMagSensorData[2] + mlxData.mlInitMagBias[2]);

    return result;
}

/**
 *  @brief  MLGetMagnetometerFloat is used to get magnetometer data
 *
 *  @pre    MLDmpOpen() \ifnot UMPL or MLDmpPedometerStandAloneOpen() \endif
 *          must have been called.
 *
 *  @param  data
 *              A pointer to an array to be passed back to the user.
 *              <b>Must be 3 cells long</b>.
 *
 *  @return ML_SUCCESS if the command is successful; an error code otherwise.
 */
tMLError MLGetMagnetometerFloat(float *data)
{
    INVENSENSE_FUNC_START;

    tMLError result = ML_SUCCESS;

    if ( MLGetState() < ML_STATE_DMP_OPENED )
        return ML_ERROR_SM_IMPROPER_STATE;

    if (NULL == data) {
        return ML_ERROR_INVALID_PARAMETER;
    }

            data[0] = (float)mlxData.mlMagCalibratedData[0]/65536.0f;
            data[1] = (float)mlxData.mlMagCalibratedData[1]/65536.0f;
            data[2] = (float)mlxData.mlMagCalibratedData[2]/65536.0f;

    return result;
}


/**
 *  @brief  MLGetPressureFloat is used to get a single value representing the pressure in Pascal
 *
 *  @pre    MLDmpOpen() \ifnot UMPL or MLDmpPedometerStandAloneOpen() \endif
 *          must have been called.
 *
 *  @param  data
 *              A pointer to the data to be passed back to the user.
 *
 *  @return ML_SUCCESS if the command is successful; an error code otherwise.
 */
tMLError MLGetPressureFloat(float *data)
{
    INVENSENSE_FUNC_START;

    tMLError result = ML_SUCCESS;

    if ( MLGetState() < ML_STATE_DMP_OPENED )
        return ML_ERROR_SM_IMPROPER_STATE;

    if (NULL == data) {
        return ML_ERROR_INVALID_PARAMETER;
    }

            data[0] = (float)mlxData.mlPressure;

    return result;
}

/**
 *  @brief  MLGetHeadingFloat is used to get single number representing the heading of the device
 *          relative to the Earth, in which 0 represents North, 90 degrees
 *          represents East, and so on.
 *          The heading is defined as the direction of the +Y axis if the Y
 *          axis is horizontal, and otherwise the direction of the -Z axis.
 *
 *  @pre    MLDmpOpen() \ifnot UMPL or MLDmpPedometerStandAloneOpen() \endif
 *          must have been called.
 *
 *  @param  data
 *              A pointer to the data to be passed back to the user.
 *
 *  @return ML_SUCCESS if the command is successful; an error code otherwise.
 */
tMLError MLGetHeadingFloat(float *data)
{
    INVENSENSE_FUNC_START;

    tMLError result = ML_SUCCESS;
    float rotMatrix[9];
	float tmp;

    if ( MLGetState() < ML_STATE_DMP_OPENED )
        return ML_ERROR_SM_IMPROPER_STATE;

    if (NULL == data) {
        return ML_ERROR_INVALID_PARAMETER;
    }

            MLGetRotationMatrixFloat(rotMatrix);
            if ((rotMatrix[7]<0.707) && (rotMatrix[7]>-0.707)) {
                tmp = (float)((double)atan2(rotMatrix[4], rotMatrix[1])*57.29577951308 - 90.0f);
            } else {
                tmp = (float)((double)atan2(rotMatrix[5], rotMatrix[2])*57.29577951308 + 90.0f);
            }
            if (tmp<0) {
                tmp += 360.0f;
            }
            data[0] = 360-tmp;


    return result;
}

/**
 *  @brief  MLGetMagBiasErrorFloat is used to get an array of three numbers representing
 *			the current estimated error in the compass biases. These numbers are unitless and serve
 *          as rough estimates in which numbers less than 100 typically represent
 *          reasonably well calibrated compass axes.
 *
 *  @pre    MLDmpOpen() \ifnot UMPL or MLDmpPedometerStandAloneOpen() \endif
 *          must have been called.
 *
 *  @param  data
 *              A pointer to an array to be passed back to the user.
 *              <b>Must be 3 cells long</b>.
 *
 *  @return ML_SUCCESS if the command is successful; an error code otherwise.
 */
tMLError MLGetMagBiasErrorFloat(float *data)
{
    INVENSENSE_FUNC_START;

    tMLError result = ML_SUCCESS;

    if ( MLGetState() < ML_STATE_DMP_OPENED )
        return ML_ERROR_SM_IMPROPER_STATE;

    if (NULL == data) {
        return ML_ERROR_INVALID_PARAMETER;
    }

            if (mlxData.mlLargeField==0) {
                data[0] = (float)mlxData.mlMagBiasError[0];
                data[1] = (float)mlxData.mlMagBiasError[1];
                data[2] = (float)mlxData.mlMagBiasError[2];
            } else {
                data[0] = (float)P_INIT;
                data[1] = (float)P_INIT;
                data[2] = (float)P_INIT;
            }

    return result;
}

/**
 *  @brief  MLGetMagScaleFloat is used to get magnetometer scale.
 *
 *  @pre    MLDmpOpen() \ifnot UMPL or MLDmpPedometerStandAloneOpen() \endif
 *          must have been called.
 *
 *  @param  data
 *              A pointer to an array to be passed back to the user.
 *              <b>Must be 3 cells long</b>.
 *
 *  @return ML_SUCCESS if the command is successful; an error code otherwise.
 */
tMLError MLGetMagScaleFloat(float *data)
{
    INVENSENSE_FUNC_START;

    tMLError result = ML_SUCCESS;

    if ( MLGetState() < ML_STATE_DMP_OPENED )
        return ML_ERROR_SM_IMPROPER_STATE;

    if (NULL == data) {
        return ML_ERROR_INVALID_PARAMETER;
    }

            data[0] = (float)mlxData.mlMagScale[0]/65536.0f;
            data[1] = (float)mlxData.mlMagScale[1]/65536.0f;
            data[2] = (float)mlxData.mlMagScale[2]/65536.0f;

    return result;
}

/**
 *  @brief  MLGetLocalFieldFloat is used to get local magnetic field data.
 *
 *  @pre    MLDmpOpen() \ifnot UMPL or MLDmpPedometerStandAloneOpen() \endif
 *          must have been called.
 *
 *  @param  data
 *              A pointer to an array to be passed back to the user.
 *              <b>Must be 3 cells long</b>.
 *
 *  @return ML_SUCCESS if the command is successful; an error code otherwise.
 */
tMLError MLGetLocalFieldFloat(float *data)
{
    INVENSENSE_FUNC_START;

    tMLError result = ML_SUCCESS;

    if ( MLGetState() < ML_STATE_DMP_OPENED )
        return ML_ERROR_SM_IMPROPER_STATE;

    if (NULL == data) {
        return ML_ERROR_INVALID_PARAMETER;
    }

            data[0] = (float)mlxData.mlLocalField[0]/65536.0f;
            data[1] = (float)mlxData.mlLocalField[1]/65536.0f;
            data[2] = (float)mlxData.mlLocalField[2]/65536.0f;

    return result;
}

/**
 *  @brief  MLGetLocalFieldFloat is used to get relative quaternion data.
 *
 *  @pre    MLDmpOpen() \ifnot UMPL or MLDmpPedometerStandAloneOpen() \endif
 *          must have been called.
 *
 *  @param  data
 *              A pointer to an array to be passed back to the user.
 *              <b>Must be 4 cells long at least</b>.
 *
 *  @return ML_SUCCESS if the command is successful; an error code otherwise.
 */
tMLError MLGetRelativeQuaternionFloat(float *data)
{
    INVENSENSE_FUNC_START;

    tMLError result = ML_SUCCESS;

    if ( MLGetState() < ML_STATE_DMP_OPENED )
        return ML_ERROR_SM_IMPROPER_STATE;

    if (NULL == data) {
        return ML_ERROR_INVALID_PARAMETER;
    }

            data[0] = (float)mlxData.mlRelativeQuat[0]/1073741824.0f;
            data[1] = (float)mlxData.mlRelativeQuat[1]/1073741824.0f;
            data[2] = (float)mlxData.mlRelativeQuat[2]/1073741824.0f;
            data[3] = (float)mlxData.mlRelativeQuat[3]/1073741824.0f;

    return result;
}

/**
 *  @brief  MLSetGyroBiasFixed is used to set Gyro Bias
 *
 *          Please refer to the provided "9-Axis Sensor Fusion
 *          Application Note" document provided.
 *
 *  @pre    MLDmpOpen() \ifnot UMPL or
 *          MLDmpPedometerStandAloneOpen() \endif
 *  @pre    MLDmpStart() must <b>NOT</b> have been called.
 *
 *  @param  data        A pointer to an array to be copied from the user.
 *
 *  @return ML_SUCCESS if successful; a non-zero error code otherwise.
 */
tMLError MLSetGyroBiasFixed(long* data)
{
    INVENSENSE_FUNC_START;
    tMLError result = ML_SUCCESS;
    long biasTmp;
    long sf = 0;
    short offset[MPU_NUM_AXES];
	int i;
	struct mldl_cfg * mldl_cfg = MLDLGetCfg();

            mlxData.mlGyroBias[0] = data[0];
            mlxData.mlGyroBias[1] = data[1];
            mlxData.mlGyroBias[2] = data[2];
            if (mldl_cfg->trim != 0) {
                sf = 2000 * 131 / mldl_cfg->trim;
            } else {
                sf = 2000;
            }
            for (i = 0; i < MPU_NUM_AXES; i++) {
                biasTmp = -mlxData.mlGyroBias[i] / sf;
                if (biasTmp < 0)
                    biasTmp += 65536L;
                offset[i] = (short)biasTmp;
            }
            result = MLDLSetOffset((const short*)offset);

    ERROR_CHECK(result);
    return ML_SUCCESS;
}

/**
 *  @brief  MLSetAccelBiasFixed is used to set Accel Bias
 *
 *          Please refer to the provided "9-Axis Sensor Fusion
 *          Application Note" document provided.
 *
 *  @pre    MLDmpOpen() \ifnot UMPL or
 *          MLDmpPedometerStandAloneOpen() \endif
 *  @pre    MLDmpStart() must <b>NOT</b> have been called.
 *
 *  @param  data        A pointer to an array to be copied from the user.
 *
 *  @return ML_SUCCESS if successful; a non-zero error code otherwise.
 */
tMLError MLSetAccelBiasFixed(long* data)
{
    INVENSENSE_FUNC_START;
    tMLError result = ML_SUCCESS;
    long biasTmp;
    long sf = 0;
    int i,j;
    unsigned char regs[6];

            mlxData.mlGyroBias[3] = data[0];
            mlxData.mlGyroBias[4] = data[1];
            mlxData.mlGyroBias[5] = data[2];
            for (i = 0; i < 3; i++) {
                if (mlxData.mlAccelSens != 0) {
                    long long tmp64 = 0;
                    for (j = 0; j < 3; j++) {
                        tmp64 += (long long)mlxData.mlGyroBias[j+3] *
                                            mlxData.mlAccelCal[i*3+j];
                    }
                    biasTmp = (long)(tmp64 / mlxData.mlAccelSens);
                    biasTmp = biasTmp * 8192 / mlxData.mlAccelSens;
                } else {
                    biasTmp = 0;
                }
                if (biasTmp < 0) biasTmp += 65536L;
                regs[2*i+0] = (unsigned char)(biasTmp / 256);
                regs[2*i+1] = (unsigned char)(biasTmp % 256);
            }
            result = MLDLSetMemoryMPU(KEY_D_1_8, 2, &regs[0]);
            ERROR_CHECK(result);
            result = MLDLSetMemoryMPU(KEY_D_1_10, 2, &regs[2]);
            ERROR_CHECK(result);
            result = MLDLSetMemoryMPU(KEY_D_1_2, 2, &regs[4]);

    ERROR_CHECK(result);
    return ML_SUCCESS;
}

/**
 *  @brief  MLSetMagBiasFixed is used to set Compass Bias
 *
 *          Please refer to the provided "9-Axis Sensor Fusion
 *          Application Note" document provided.
 *
 *  @pre    MLDmpOpen() \ifnot UMPL or
 *          MLDmpPedometerStandAloneOpen() \endif
 *  @pre    MLDmpStart() must <b>NOT</b> have been called.
 *
 *  @param  data        A pointer to an array to be copied from the user.
 *
 *  @return ML_SUCCESS if successful; a non-zero error code otherwise.
 */
tMLError MLSetMagBiasFixed(long* data)
{
    INVENSENSE_FUNC_START;
    tMLError result = ML_SUCCESS;

            CompassSetBias( data );
            mlxData.mlInitMagBias[0] = 0;
            mlxData.mlInitMagBias[1] = 0;
            mlxData.mlInitMagBias[2] = 0;
            mlxData.mlGotCompassBias = 1;
            mlxData.mlGotInitCompassBias = 1;
            mlxData.mlCompassState = SF_STARTUP_SETTLE;

    ERROR_CHECK(result);
    return ML_SUCCESS;
}

/**
 *  @brief  MLSetGyroTempSlopeFixed is used to set Gyro temperature slope
 *
 *          Please refer to the provided "9-Axis Sensor Fusion
 *          Application Note" document provided.
 *
 *  @pre    MLDmpOpen() \ifnot UMPL or
 *          MLDmpPedometerStandAloneOpen() \endif
 *  @pre    MLDmpStart() must <b>NOT</b> have been called.
 *
 *  @param  data        A pointer to an array to be copied from the user.
 *
 *  @return ML_SUCCESS if successful; a non-zero error code otherwise.
 */
tMLError MLSetGyroTempSlopeFixed(long* data)
{
    INVENSENSE_FUNC_START;
    tMLError result = ML_SUCCESS;
	int i;
	long sf;
	unsigned char regs[3];

            mlxData.mlFactoryTempComp = 1;
            mlxData.mlTempSlope[0] = data[0];
            mlxData.mlTempSlope[1] = data[1];
            mlxData.mlTempSlope[2] = data[2];
            for (i = 0; i < MPU_NUM_AXES; i++) {
                sf = -mlxData.mlTempSlope[i] / 1118;
                if (sf>127) {
                    sf-=256;
                }
                regs[i] = (unsigned char)sf;
            }
            result = MLDLSetOffsetTC(regs);

    ERROR_CHECK(result);
    return ML_SUCCESS;
}

/**
 *  @brief  MLSetLocalFieldFixed is used to set local magnetic field
 *
 *          Please refer to the provided "9-Axis Sensor Fusion
 *          Application Note" document provided.
 *
 *  @pre    MLDmpOpen() \ifnot UMPL or
 *          MLDmpPedometerStandAloneOpen() \endif
 *  @pre    MLDmpStart() must <b>NOT</b> have been called.
 *
 *  @param  data        A pointer to an array to be copied from the user.
 *
 *  @return ML_SUCCESS if successful; a non-zero error code otherwise.
 */
tMLError MLSetLocalFieldFixed(long* data)
{
    INVENSENSE_FUNC_START;
    tMLError result = ML_SUCCESS;

            mlxData.mlLocalField[0] = data[0];
            mlxData.mlLocalField[1] = data[1];
            mlxData.mlLocalField[2] = data[2];
            mlxData.mlNewLocalField = 1;

    ERROR_CHECK(result);
    return ML_SUCCESS;
}

/**
 *  @brief  MLSetMagScaleFixed is used to set magnetometer scale
 *
 *          Please refer to the provided "9-Axis Sensor Fusion
 *          Application Note" document provided.
 *
 *  @pre    MLDmpOpen() \ifnot UMPL or
 *          MLDmpPedometerStandAloneOpen() \endif
 *  @pre    MLDmpStart() must <b>NOT</b> have been called.
 *
 *  @param  data        A pointer to an array to be copied from the user.
 *
 *  @return ML_SUCCESS if successful; a non-zero error code otherwise.
 */
tMLError MLSetMagScaleFixed(long* data)
{
    INVENSENSE_FUNC_START;
    tMLError result = ML_SUCCESS;

            mlxData.mlMagScale[0] = data[0];
            mlxData.mlMagScale[1] = data[1];
            mlxData.mlMagScale[2] = data[2];

    ERROR_CHECK(result);
    return ML_SUCCESS;
}

/**
 *  @brief  MLSetGyroTempSlopeFloat is used to set gyro temperature slope
 *
 *          Please refer to the provided "9-Axis Sensor Fusion
 *          Application Note" document provided.
 *
 *  @pre    MLDmpOpen() \ifnot UMPL or
 *          MLDmpPedometerStandAloneOpen() \endif
 *  @pre    MLDmpStart() must <b>NOT</b> have been called.
 *
 *  @param  data        A pointer to an array to be copied from the user.
 *
 *  @return ML_SUCCESS if successful; a non-zero error code otherwise.
 */
tMLError MLSetGyroTempSlopeFloat(float * data)
{
    long arrayTmp[3];
    arrayTmp[0] = (long)(data[0]*65536.f);
    arrayTmp[1] = (long)(data[1]*65536.f);
    arrayTmp[2] = (long)(data[2]*65536.f);
    return MLSetGyroTempSlopeFixed(arrayTmp);
}

/**
 *  @brief  MLSetGyroBiasFloat is used to set gyro bias
 *
 *          Please refer to the provided "9-Axis Sensor Fusion
 *          Application Note" document provided.
 *
 *  @pre    MLDmpOpen() \ifnot UMPL or
 *          MLDmpPedometerStandAloneOpen() \endif
 *  @pre    MLDmpStart() must <b>NOT</b> have been called.
 *
 *  @param  data        A pointer to an array to be copied from the user.
 *
 *  @return ML_SUCCESS if successful; a non-zero error code otherwise.
 */
tMLError MLSetGyroBiasFloat(float * data)
{
    long arrayTmp[3];
    arrayTmp[0] = (long)(data[0]*65536.f);
    arrayTmp[1] = (long)(data[1]*65536.f);
    arrayTmp[2] = (long)(data[2]*65536.f);
    return MLSetGyroBiasFixed(arrayTmp);

}

/**
 *  @brief  MLSetAccelBiasFloat is used to set accel bias
 *
 *          Please refer to the provided "9-Axis Sensor Fusion
 *          Application Note" document provided.
 *
 *  @pre    MLDmpOpen() \ifnot UMPL or
 *          MLDmpPedometerStandAloneOpen() \endif
 *  @pre    MLDmpStart() must <b>NOT</b> have been called.
 *
 *  @param  data        A pointer to an array to be copied from the user.
 *
 *  @return ML_SUCCESS if successful; a non-zero error code otherwise.
 */
tMLError MLSetAccelBiasFloat(float * data)
{
    long arrayTmp[3];
    arrayTmp[0] = (long)(data[0]*65536.f);
    arrayTmp[1] = (long)(data[1]*65536.f);
    arrayTmp[2] = (long)(data[2]*65536.f);
    return MLSetAccelBiasFixed(arrayTmp);

}

/**
 *  @brief  MLSetMagBiasFloat is used to set compass bias
 *
 *          Please refer to the provided "9-Axis Sensor Fusion
 *          Application Note" document provided.
 *
 *  @pre    MLDmpOpen()\ifnot UMPL or
 *          MLDmpPedometerStandAloneOpen() \endif
 *  @pre    MLDmpStart() must <b>NOT</b> have been called.
 *
 *  @param  data        A pointer to an array to be copied from the user.
 *
 *  @return ML_SUCCESS if successful; a non-zero error code otherwise.
 */
tMLError MLSetMagBiasFloat(float * data)
{
    long arrayTmp[3];
    arrayTmp[0] = (long)(data[0]*65536.f);
    arrayTmp[1] = (long)(data[1]*65536.f);
    arrayTmp[2] = (long)(data[2]*65536.f);
    return MLSetMagBiasFixed(arrayTmp);
}

/**
 *  @brief  MLSetLocalFieldFloat is used to set local magnetic field
 *
 *          Please refer to the provided "9-Axis Sensor Fusion
 *          Application Note" document provided.
 *
 *  @pre    MLDmpOpen() \ifnot UMPL or
 *          MLDmpPedometerStandAloneOpen() \endif
 *  @pre    MLDmpStart() must <b>NOT</b> have been called.
 *
 *  @param  data        A pointer to an array to be copied from the user.
 *
 *  @return ML_SUCCESS if successful; a non-zero error code otherwise.
 */
tMLError MLSetLocalFieldFloat(float * data)
{
    long arrayTmp[3];
    arrayTmp[0] = (long)(data[0]*65536.f);
    arrayTmp[1] = (long)(data[1]*65536.f);
    arrayTmp[2] = (long)(data[2]*65536.f);
    return MLSetLocalFieldFixed(arrayTmp);
}

/**
 *  @brief  MLSetMagScaleFloat is used to set magnetometer scale
 *
 *          Please refer to the provided "9-Axis Sensor Fusion
 *          Application Note" document provided.
 *
 *  @pre    MLDmpOpen() \ifnot UMPL or
 *          MLDmpPedometerStandAloneOpen() \endif
 *  @pre    MLDmpStart() must <b>NOT</b> have been called.
 *
 *  @param  data        A pointer to an array to be copied from the user.
 *
 *  @return ML_SUCCESS if successful; a non-zero error code otherwise.
 */
tMLError MLSetMagScaleFloat(float * data)
{
    long arrayTmp[3];
    arrayTmp[0] = (long)(data[0]*65536.f);
    arrayTmp[1] = (long)(data[1]*65536.f);
    arrayTmp[2] = (long)(data[2]*65536.f);
    return MLSetMagScaleFixed(arrayTmp);
}


