/*
 $License:
    Copyright (C) 2011 InvenSense Corporation, All Rights Reserved.
 $
 */

/******************************************************************************
 *
 * $Id:$
 *
 *****************************************************************************/

#include <stdio.h>
#include <string.h>

#include "log.h"
#undef MPL_LOG_TAG
#define MPL_LOG_TAG "ml-tempcomp"

#include "tempComp.h"
#include "progressiveNoMotion.h"
#include "gyroBias.h"
#include "ml.h"
#include "mlFIFO.h"
#include "mlmath.h"
#include "mlsupervisor.h"
#include "mlMathFunc.h"

#define TEMPCOMP_DEBUG 0

#define TEMPCOMP_LOG(...)  \
	if (TEMPCOMP_DEBUG) \
		MPL_LOGI(__VA_ARGS__);

/*
    Data Structures
*/
struct _TC {
    short haveSlope;
    short haveTempOffset;
    int noMotionTimer;
    long motionTimer;
    float noMotionBiases[3];
    float noMotionTemp;
    long prevBias[3];
    float lastTemp;
    float lastGyroData[3];
    int gotLastData;
};


/*
    Globals 
*/
extern tMLSupervisorCB ml_supervisor_cb;

struct _TC tcData;

/*
    Prototypes
*/
tMLError MLTempCompSupervisor(unsigned long deltaTime);

/*
    Functions
*/

/**
 *  @brief  Enable the temperature compensation algorithm and calibrated gyro 
 *          temperature compensated output.
 *  @return ML_SUCCESS == 0.
 */
tMLError MLEnableTempComp(void)
{
	TEMPCOMP_LOG("ENABLE\n")
	
    ml_supervisor_cb.MLTempCompSupervisor = MLTempCompSupervisor;
    (void) GyroBiasRegisterTracker(1, GYROBIAS_TRACKER_TEMPCOMP);
    return ML_SUCCESS;
}

/**
 *  @brief  Disable the temperature compensation algorithm and calibrated gyro 
 *          temperature compensated output.
 *  @return ML_SUCCESS == 0.
 */
tMLError MLDisableTempComp(void)
{
	TEMPCOMP_LOG("DISABLE\n")
	
    ml_supervisor_cb.MLTempCompSupervisor = NULL;
    (void) GyroBiasRegisterTracker(0, GYROBIAS_TRACKER_TEMPCOMP);
    return ML_SUCCESS;
}


/**
 *  @brief  Say if the temperature compensation algorithm is fully active.
 *  @return 1 if the temp co is enabled. 0 if it is not.
 */
int TempCompIsEnabled(void)
{
    return (
        (mlParams.biasUpdateFunc & ML_LEARN_BIAS_FROM_TEMPERATURE) 
            &&
        (ml_supervisor_cb.MLTempCompSupervisor != NULL)
            &&
        MLGetGyroPresent()
    );
}

/**
 *  @brief  Reset the temperature compensation algorithm internal state 
 *          machine.
 */
void TempCompReset(void)
{
	TEMPCOMP_LOG("RESET\n")
    memset(&tcData, 0, sizeof(tcData));
        
    tcData.motionTimer = 1000;
    tcData.gotLastData = TRUE;
}

/**
 *  @brief  Whether the temperature compensation slope has been already 
 *          computed.
 */
int TempCompHaveSlope(void)
{
    return (tcData.haveSlope);
}

/**
 *  @brief  Whether the temperature compensation offset have been 
 *          calculated off the slope and the current temperature at
 *          least once.
 */
int TempCompHaveOffset(void)
{
    return (tcData.haveTempOffset);
}


/**
 *  @brief  Get the temperature change from last and previous time
 *          TempCompApply has executed.
 *          This is a sinple shorthand to avoid making the internal
 *          data structure struct _TC non file static.
 *  @return The temperature difference in degrees C.
 */
float TempCompGetTempDifference(void)
{
    return (tcData.lastTemp - tcData.noMotionTemp);
}


/**
 *  @internal
 *  @brief  Get the temperature in float format. 
 *          Represented in degree C.
 *  @param  temp
 *              a pointer to store the temperature.
 *  @return ML_SUCCESS if successful, a non-zero error code otherwise.
 */
static tMLError TempCompGetTemp(float *temp)
{
    long temperature;
    tMLError result;

    result = FIFOGetTemperature(&temperature);
    //ERROR_CHECK(result);  // FIXME : should return an error, for now we don't
    *temp = ((float)temperature) / 65536.0f;
    return ML_SUCCESS;
}

/**
 *  @internal
 *  @brief  Get raw gyro data, transform to float, scale, and add in bias.
 *          Expressed in degree per second.
 *  @param  gyro
 *              A pointer to store the gyro data for the 3 axis.
 *  @return ML_SUCCESS if successful, a non-zero error code otherwise.
 */
static tMLError TempCompGetGyro(float gyro[3])
{
    long sensors[6];
    int i;
    tMLError result;

    result = FIFOGetSensorGyroData(sensors);
    //ERROR_CHECK(result);  // FIXME : should return an error, for now we don't
    for (i = 0; i < GYRO_NUM_AXES; i++) {
        gyro[i] =  (float)sensors[i] * mlxData.mlGyroSens / 1073741824.0f;
        gyro[i] += (float)mlxData.mlGyroBias[i] / 65536.0f;
    }
    return ML_SUCCESS;
}

/**
 *  @brief  Find the right temperature bin.
 *  @param  temp
 *              The temperature in degree C.
 *  @return the temperature bin number, [0, BINS).
 */
int TempCompFindTempBin(float temp)
{
    int bin;

    bin = (int)((temp - MIN_TEMP) / ((MAX_TEMP - MIN_TEMP) / BINS));
    if (bin < 0)
        bin = 0;
    if (bin > BINS - 1)
        bin = BINS - 1;
    return bin;
}

/**
 *  @internal
 *  @brief  Find the closest available temperature bin (typically 5 C wide) to 
 *          the requested bin.  For a bin to be selected, it has to be either
 *          be declared valid (mlTempValidData == 1) or it's got at least 
 *          PTS_PER_BIN points in the temp. compensation table.
 *  @return the temperature bin number, [0, BINS).
 */
int TempCompFindClosestBin(int bin)
{
    int inc = 1;
    int incSign = 0;

    while (!mlxData.mlTempValidData[bin]
                && mlxData.mlTempPtrs[bin] == 0
                && inc < 4) {
        if (incSign == 0) {
            bin += inc;
            incSign = 1;
        } else {
            bin -= inc;
            incSign = 0;
        }
        inc += 1;
    }
    return (inc < 4 ? bin : -1);
}


/**
 *  @brief  Prints the temperature compesation table for debugging 
 *          purpose.
 */
void TempCompPrintTable(void)
{
#define STR_LEN 500
    char line[STR_LEN];
    int k, j;

    MPL_LOGI("\n");
    MPL_LOGI("mlTempData & mlTempPtrs\n");
    for (j = 0, line[0] = 0; j < BINS; j++) {
        sprintf(line, "pnt %2d - ", j);
        for (k = 0; k < PTS_PER_BIN; k++) {
            sprintf(line, "%s%+9.3f ", line, mlxData.mlTempData[j][k]);
        }
        sprintf(line, "%s - %d", line, mlxData.mlTempPtrs[j]);
        MPL_LOGI("%s\n", line);
    }

    MPL_LOGI("mlXGyroTempData\n");
    for (j = 0, line[0] = 0; j < BINS; j++) {
        sprintf(line, "pnt %2d - ", j);
        for (k = 0; k < PTS_PER_BIN; k++) {
            sprintf(line, "%s%+9.3f ", line, mlxData.mlXGyroTempData[j][k]);
        }
        MPL_LOGI("%s\n", line);
    }
    MPL_LOGI("mlYGyroTempData\n");
    for (j = 0, line[0] = 0; j < BINS; j++) {
        sprintf(line, "pnt %2d - ", j);
        for (k = 0; k < PTS_PER_BIN; k++) {
            sprintf(line, "%s%+9.3f ", line, mlxData.mlYGyroTempData[j][k]);
        }
        MPL_LOGI("%s\n", line);
    }
    MPL_LOGI("mlZGyroTempData\n");
    for (j = 0, line[0] = 0; j < BINS; j++) {
        sprintf(line, "pnt %2d - ", j);
        for (k = 0; k < PTS_PER_BIN; k++) {
            sprintf(line, "%s%+9.3f ", line, mlxData.mlZGyroTempData[j][k]);
        }
        MPL_LOGI("%s\n", line);
    }
    MPL_LOGI("\n");
}

/**
 *  @brief  Add old data to temperature table. 
 *          Adding current data risks adding a motion point, since there is a 
 *          1 second delay between motion and the polling based motion detect.
 *  @return ML_SUCCESS if successful, a non-zero error code otherwise.
 */
static tMLError TempCompAddData(void)
{
    float newGyroData[3];
    float newTemp;
    int i;

    /* since AddData runs only in "no motion" state, this point represents
       a no motion point */
    TempCompGetTemp(&newTemp);
    TempCompGetGyro(newGyroData);

    /* got data from a previous run */
    if (tcData.gotLastData) { 
        int bin = TempCompFindTempBin(tcData.lastTemp);
        int *tempPtr = &(mlxData.mlTempPtrs[bin]);

        mlxData.mlTempData[bin][*tempPtr] = tcData.lastTemp;
        mlxData.mlXGyroTempData[bin][*tempPtr] = tcData.lastGyroData[0];
        mlxData.mlYGyroTempData[bin][*tempPtr] = tcData.lastGyroData[1];
        mlxData.mlZGyroTempData[bin][*tempPtr] = tcData.lastGyroData[2];

        if (TEMPCOMP_DEBUG) {
            MPL_LOGI(
                "TempComp : add data to temp table : "
                "%+10.3f degC - %+10.3f %+10.3f %+10.3f dps\n",
                tcData.lastTemp,
                tcData.lastGyroData[0],
                tcData.lastGyroData[1],
                tcData.lastGyroData[2]
                );
            TempCompPrintTable();
        }

        /* Treat each bin as a circular buffer 
           If pointer wraps around, set the data valid bit, 
           indicating that the bin is full. */
        *tempPtr = (*tempPtr + 1) % PTS_PER_BIN;
        if (*tempPtr == 0) 
            mlxData.mlTempValidData[bin] = TRUE;

        /*  Track the current bias and temperature to be used as the offset 
            (temp comp table will only be used to generate the slope) */
        for (i = 0; i < GYRO_NUM_AXES; i++) {
            tcData.noMotionBiases[i] = (float)mlxData.mlGyroBias[i] / 65536.0f;
        }
        tcData.noMotionTemp = tcData.lastTemp;
    }

    /*  if the biases have not changed since last call to this function,
        save current temperature and gyro data to be used next time;
        see above the use of lastTemp and lastGyroData */
    if (tcData.prevBias[0] == mlxData.mlGyroBias[0] && 
        tcData.prevBias[1] == mlxData.mlGyroBias[1] &&
        tcData.prevBias[2] == mlxData.mlGyroBias[2]) {
        tcData.lastTemp = newTemp;
        for (i = 0; i < 3; i++) {
            tcData.lastGyroData[i] = newGyroData[i];
        }
        tcData.gotLastData = TRUE;
    } else {
        tcData.gotLastData = FALSE;
    }

    /* save the current biases at this point to be 
       able, at the next run, to verify they haven't changed
       (see block above) */
    tcData.prevBias[0] = mlxData.mlGyroBias[0];
    tcData.prevBias[1] = mlxData.mlGyroBias[1];
    tcData.prevBias[2] = mlxData.mlGyroBias[2];
    return ML_SUCCESS;
}

/**
 *  @brief  Walk through temperature table and construct a linear fit.
 *          Should only access bins near current temperature?
 *  @return ML_SUCCESS if successful, a non-zero error code otherwise.
 */
static tMLError TempCompRecompute(void)
{
#define ORD 1
    float mXGyro[3] = {0};
    float mYGyro[3] = {0};
    float mZGyro[3] = {0};
    float m[3][3] = {{0}};
    float minv[3][3] = {{0}};
    float det = 0;
    int i, j, k;

    for (j = 0; j < BINS; j++) {
        for (k = 0; k < PTS_PER_BIN; k++) {
            float temp;
            if (!mlxData.mlTempValidData[j]
                    && k >= mlxData.mlTempPtrs[j]) {
                continue;
            }

            temp = mlxData.mlTempData[j][k];

            m[0][0] += 1.f;
            m[0][1] += temp;
            m[0][2] += temp * temp;
            m[1][0] += temp;
            m[1][1] += temp * temp;
            m[1][2] += temp * temp * temp;
            m[2][0] += temp * temp;
            m[2][1] += temp * temp * temp;
            m[2][2] += temp * temp * temp * temp;

            mXGyro[0] += mlxData.mlXGyroTempData[j][k];
            mXGyro[1] += mlxData.mlXGyroTempData[j][k] * temp;
            mXGyro[2] += mlxData.mlXGyroTempData[j][k] * temp * temp;

            mYGyro[0] += mlxData.mlYGyroTempData[j][k];
            mYGyro[1] += mlxData.mlYGyroTempData[j][k] * temp;
            mYGyro[2] += mlxData.mlYGyroTempData[j][k] * temp * temp;

            mZGyro[0] += mlxData.mlZGyroTempData[j][k];
            mZGyro[1] += mlxData.mlZGyroTempData[j][k] * temp;
            mZGyro[2] += mlxData.mlZGyroTempData[j][k] * temp * temp;
        }
    }

#if (ORD == 2)
    /* Second order fit (not currently used, probably better to 
       do piece-wise linear) */
    det = 
        m[0][0] * m[1][1] * m[2][2] - m[0][0] * m[1][2] * m[2][1] -
        m[0][1] * m[1][0] * m[2][2] + m[0][1] * m[1][2] * m[2][0] +
        m[0][2] * m[1][0] * m[2][1] - m[0][2] * m[1][1] * m[2][0];

    /* check that determinant is big enough */
    if (fabs(det) < 100.f) {
        return;
    } 

    /* Matrix inversion for second order fit */
    minv[0][0] = (m[1][1] * m[2][2] - m[1][2] * m[2][1]) / det;
    minv[0][1] = (m[0][2] * m[2][1] - m[0][1] * m[2][2]) / det;
    minv[0][2] = (m[0][1] * m[1][2] - m[0][2] * m[1][1]) / det;
    minv[1][0] = (m[1][2] * m[2][0] - m[1][0] * m[2][2]) / det;
    minv[1][1] = (m[0][0] * m[2][2] - m[0][2] * m[2][0]) / det;
    minv[1][2] = (m[0][2] * m[1][0] - m[0][0] * m[1][2]) / det;
    minv[2][0] = (m[1][0] * m[2][1] - m[1][1] * m[2][0]) / det;
    minv[2][1] = (m[0][1] * m[2][0] - m[0][0] * m[2][1]) / det;
    minv[2][2] = (m[0][0] * m[1][1] - m[0][1] * m[1][0]) / det;

    /* Populate coefficients for second order fit */
    for (i = 0; i < 3; i++) {
        mlxData.mlXGyroCoeff[i] = 0.f;
        mlxData.mlYGyroCoeff[i] = 0.f;
        mlxData.mlZGyroCoeff[i] = 0.f;
        for (j = 0; j < 3; j++) {
            mlxData.mlXGyroCoeff[i] += minv[j][i] * mXGyro[j];
            mlxData.mlYGyroCoeff[i] += minv[j][i] * mYGyro[j];
            mlxData.mlZGyroCoeff[i] += minv[j][i] * mZGyro[j];
        }
    }
#else
    /* Linear fit */
    det = m[0][0] * m[1][1] - m[0][1] * m[1][0];

    /* check that determinant is big enough;
       (temperature range has to be at least 8 deg. C wide) */
    if (fabs(det) < 100.f) {
        return ML_SUCCESS;
    } 

    /* Matrix inversion for linear fit */
    minv[0][0] = +m[1][1] / det;
    minv[0][1] = -m[1][0] / det;
    minv[1][0] = -m[0][1] / det;
    minv[1][1] = +m[0][0] / det;

    /* Populate coefficients for first order fit */
    for (i = 0; i < 2; i++) {
        mlxData.mlXGyroCoeff[i] = 0.f;
        mlxData.mlYGyroCoeff[i] = 0.f;
        mlxData.mlZGyroCoeff[i] = 0.f;
        for (j = 0; j < 2; j++) {
            mlxData.mlXGyroCoeff[i] += minv[j][i] * mXGyro[j];
            mlxData.mlYGyroCoeff[i] += minv[j][i] * mYGyro[j];
            mlxData.mlZGyroCoeff[i] += minv[j][i] * mZGyro[j];
        }
    }
#endif

    tcData.haveSlope = TRUE;

    if (TEMPCOMP_DEBUG)
        MPL_LOGI(
            "TempComp : recomputed temp comp slopes : "
            "%+10.3f %+10.3f %+10.3f dps/C\n",
            mlxData.mlXGyroCoeff[1],
            mlxData.mlYGyroCoeff[1],
            mlxData.mlZGyroCoeff[1]);
    return ML_SUCCESS;
}

/**
 *  @brief  Apply temperature compensation table to gyro bias.
 */
static void TempCompApply(void)
{
    float newTemp;

    tcData.gotLastData = FALSE;

    TempCompGetTemp(&newTemp);

	if (tcData.haveSlope) {
		TEMPCOMP_LOG("APPLY: temp comp have slope\n")
	} else { 
		TEMPCOMP_LOG("APPLY: temp comp no slope\n")
	}

    if (tcData.haveSlope) {
        const float tempDelta = newTemp - tcData.noMotionTemp;
        /*  Use slope from temperature table, and 
            offset from most recent no motion bias to prevent 
            any hysteresis effects */
        mlxData.mlGyroTempBias[0] = (long)(
            (tcData.noMotionBiases[0] + mlxData.mlXGyroCoeff[1] * tempDelta) * 
            65536.0f);
        mlxData.mlGyroTempBias[1] = (long)(
            (tcData.noMotionBiases[1] + mlxData.mlYGyroCoeff[1] * tempDelta) * 
            65536.0f);
        mlxData.mlGyroTempBias[2] = (long)(
            (tcData.noMotionBiases[2] + mlxData.mlZGyroCoeff[1] * tempDelta) * 
            65536.0f);
        mlxData.mlGotNoMotionBias = TRUE;
   
        /* signal the bias tracker the gyro temp comp offset are
           ready to use */
        tcData.haveTempOffset = TRUE; 

        if (TEMPCOMP_DEBUG) 
            MPL_LOGI(
                "TempComp : slope applied : %+10.3f degC\n"
                "(%+10.3f + %+10.3f) = %+10.3f\n"
                "(%+10.3f + %+10.3f) = %+10.3f\n"
                "(%+10.3f + %+10.3f) = %+10.3f\n",
                tempDelta,
                tcData.noMotionBiases[0], mlxData.mlXGyroCoeff[1] * tempDelta,
                (float)mlxData.mlGyroTempBias[0] / 65536.f,
                tcData.noMotionBiases[1], mlxData.mlYGyroCoeff[1] * tempDelta,
                (float)mlxData.mlGyroTempBias[1] / 65536.f,
                tcData.noMotionBiases[2], mlxData.mlZGyroCoeff[1] * tempDelta,
                (float)mlxData.mlGyroTempBias[2] / 65536.f);
    }
}

/**
 *  @brief  Recompute the temperature compensation table using the values
 *          loaded from file.
 *          Apply the value by finding the best guess for the gyro offsets.
 *  @return ML_SUCCESS if successful, a non-zero error code otherwise.
 */
tMLError TempCompLoadCalibrationHandler(void)
{
    float newBiases[3];
    float newTemp;
    int bin;
    tMLError result;
    int i;
	TEMPCOMP_LOG("Load Calibration Handler\n")

    tcData.gotLastData = FALSE;

    /* recompute the temp comp table */
    result = TempCompRecompute();
    ERROR_CHECK(result);

    result = TempCompGetTemp(&newTemp);
    ERROR_CHECK(result);

    /* find best guess for gyro offset */
    bin = TempCompFindTempBin(newTemp);

    /* If we have a slope from the ORD order fit, 
        apply it to the biases with B = a + b T */
    if (tcData.haveSlope) {
        newBiases[0] = 
            mlxData.mlXGyroCoeff[0] + mlxData.mlXGyroCoeff[1] * newTemp;
        newBiases[1] = 
            mlxData.mlYGyroCoeff[0] + mlxData.mlYGyroCoeff[1] * newTemp;
        newBiases[2] = 
            mlxData.mlZGyroCoeff[0] + mlxData.mlZGyroCoeff[1] * newTemp;
        MLSetGyroBiasFloat( newBiases);

        tcData.noMotionTemp = newTemp;
        for (i = 0; i < GYRO_NUM_AXES; i++) {
            tcData.noMotionBiases[i] = newBiases[i];
            mlxData.mlGyroTempBias[i] = mlxData.mlGyroBias[i];
        }

        mlxData.mlGotNoMotionBias = TRUE;  // FIXME ? declare no motion if there is more than 5 deg C difference ?

    } else {

        /* set the bias to one from the closest bin */
        int ptr;
        bin = TempCompFindClosestBin(bin);
        if (bin < 0)
            return ML_SUCCESS;
        if (mlxData.mlTempPtrs[bin] == 0) 
            ptr = PTS_PER_BIN - 1;
        else
            ptr = mlxData.mlTempPtrs[bin] - 1;

        newBiases[0] = mlxData.mlXGyroTempData[bin][ptr];
        newBiases[1] = mlxData.mlYGyroTempData[bin][ptr];
        newBiases[2] = mlxData.mlZGyroTempData[bin][ptr];
		TEMPCOMP_LOG("No slope, setting gyro bias from closest bin "
				     " %4.3f %4.3f %4.3f\n",newBiases[0],newBiases[1],newBiases[2])
        MLSetGyroBiasFloat( newBiases);

        /* record reference temp and biases */
        tcData.noMotionTemp = mlxData.mlTempData[bin][ptr];
        for (i = 0; i < GYRO_NUM_AXES; i++) {
            tcData.noMotionBiases[i] = newBiases[i];
            /* instead of loading it on mlxData.mlGyroTempBias, we assign it
               to mlxData.mlNoMotionBias to mimick having no motion biases 
               (more short term biases than a temp comp correction */
            mlxData.mlNoMotionBias[i] = mlxData.mlGyroBias[i]; 
        }
        /* declare no motion only if the temperature change is not bigger
           than 5 deg. C */
        if (fabs(newTemp - tcData.noMotionTemp) < 5.f) {
            mlxData.mlGotNoMotionBias = TRUE;
        }
    }

	TempCompApply();
    return ML_SUCCESS;
}

/**
 *  @brief  Main entry point of the temperature compensation algorithm.
 *  @param  deltaTime
 *              The time elapsed since the last call.
 *  @return ML_SUCCESS == 0.
 */
tMLError MLTempCompSupervisor(unsigned long deltaTime)
{    
    tMLError result;

    /* FIXME : remove MLSetBiasUpdateFunc enable (ML_LEARN_BIAS_FROM_TEMPERATURE)
               and then get rid of this check because this supervisor should
               not be called unless MLEnableTempComp has been called */
    if (mlxData.calLoadedFlag) {
		mlxData.calLoadedFlag = FALSE;
		result = TempCompLoadCalibrationHandler();
	}
	
	if (!TempCompIsEnabled()) {
        return ML_SUCCESS;
    }

    if (ProgressiveNoMotionGetState() == PROG_NO_MOTION) {                   /* no motion */
        tcData.noMotionTimer += deltaTime;
        if (tcData.noMotionTimer > 1000) {
			TEMPCOMP_LOG("Supervisor: no motion\n")
            tcData.noMotionTimer = 0;
            result = TempCompAddData();
            ERROR_CHECK(result);
            result = TempCompRecompute();
            ERROR_CHECK(result);
        }
    } else {                                                                 /* motion */
        tcData.motionTimer += deltaTime;
        if (tcData.motionTimer > 1200) {
			TEMPCOMP_LOG("Supervisor: motion\n")  
            tcData.motionTimer = 0;
			TempCompApply();
        }
    }

    return ML_SUCCESS;
}


/**
 *  @}
 */
