/*
 $License:
    Copyright (C) 2010 InvenSense Corporation, All Rights Reserved.
 $
 */

#ifndef __MLSL_H__
#define __MLSL_H__

/**
 *  @defgroup MLSL
 *  @brief    Motion Library - Serial Layer.
 *	      The Motion Library System Layer provides the Motion Library the
 *	      interface to the system functions.
 *
 *  @{
 *	      @file   mlsl.h
 *	      @brief  The Motion Library System Layer.
 *
 */

/*
  NOTE : this document is best viewed with 8 spaces per tab.
*/

#include "mltypes.h"
#include "mpu.h"

#ifdef __cplusplus
extern "C" {
#endif

/* ------------ */
/* - Defines. - */
/* ------------ */

/*
 * NOTE : to properly support Yamaha compass reads,
 *	  the max transfer size should be at least 9 B.
 *	  Length in bytes, typically a power of 2 >= 2
 */
#define SERIAL_MAX_TRANSFER_SIZE 128

/* ---------------------- */
/* - Types definitions. - */
/* ---------------------- */

/* --------------------- */
/* - Function p-types. - */
/* --------------------- */

/**
 *  @brief  used to open the serial port.
 *	    This port is used to send and receive data to the device.
 *
 *  @note   This function is called by MLSerialOpen().
 *	    Unlike previous MPL Software releases, explicitly calling
 *	    MLSerialOpen() is mandatory to instantiate the communication
 *	    with the device.
 *
 *  @param  port
 *		The COM port specification associated with the device in use.
 *  @param  sl_handle
 *		a pointer to the file handle to the serial device to be open
 *		for the communication.
 *
 *  @return ML_SUCCESS if successful, a non-zero error code otherwise.
 */
tMLError MLSLSerialOpen(char const *port, void **sl_handle);

/**
 *  @brief  used to close the serial port.
 *	    This port is used to send and receive data to the device.
 *
 *  @note   This function is called by MLSerialClose().
 *	    Unlike previous MPL Software releases, explicitly calling
 *	    MLSerialClose() is mandatory to properly shut-down the
 *	    communication with the device.
 *
 *  @param  sl_handle
 *		a file handle to the serial device used for the communication.
 *
 *  @return ML_SUCCESS if successful, a non-zero error code otherwise.
 */
tMLError MLSLSerialClose(void *sl_handle);

/**
 *  @brief  used to reset any buffering the driver may be doing
 *  @return ML_SUCCESS if successful, a non-zero error code otherwise.
 */
tMLError MLSLSerialReset(void *sl_handle);

/**
 *  @brief  used to write a single byte of data.
 *	    It is called by the MPL to write a single byte of data to the MPU.
 *	    This should be sent by I2C or SPI.
 *
 *  @param  sl_handle
 *		a file handle to the serial device used for the communication.
 *  @param  slaveAddr
 *		I2C slave address of device.
 *  @param  registerAddr
 *		Register address to write.
 *  @param  data
 *		Single byte of data to write.
 *
 *  @return ML_SUCCESS if successful, a non-zero error code otherwise.
 */
tMLError MLSLSerialWriteSingle(void *sl_handle,
			       unsigned char slaveAddr,
			       unsigned char registerAddr,
			       unsigned char data);

/**
 *  @brief  used to write multiple bytes of data to registers.
 *	    This should be sent by I2C or SPI.
 *
 *  @param  sl_handle
 *		a file handle to the serial device used for the communication.
 *  @param  slaveAddr
 *		I2C slave address of device.
 *  @param  registerAddr
 *		Register address to write.
 *  @param  length
 *		Length of burst of data.
 *  @param  data
 *		Pointer to block of data.
 *
 *  @return ML_SUCCESS if successful, a non-zero error code otherwise.
 */
 tMLError MLSLSerialWrite(void *sl_handle,
			 unsigned char slaveAddr,
			 unsigned short length,
			 unsigned char const *data);

/**
 *  @brief  used to read multiple bytes of data from registers.
 *	    This should be sent by I2C or SPI.
 *
 *  @param  sl_handle
 *		a file handle to the serial device used for the communication.
 *  @param  slaveAddr
 *		I2C slave address of device.
 *  @param  registerAddr
 *		Register address to read.
 *  @param  length
 *		Length of burst of data.
 *  @param  data
 *		Pointer to block of data.
 *
 *  @return ML_SUCCESS == 0 if successful; a non-zero error code otherwise.
 */
tMLError MLSLSerialRead(void *sl_handle,
			unsigned char slaveAddr,
			unsigned char registerAddr,
			unsigned short length,
			unsigned char *data);

/**
 *  @brief  used to read multiple bytes of data from the memory.
 *	    This should be sent by I2C or SPI.
 *
 *  @param  sl_handle
 *		a file handle to the serial device used for the communication.
 *  @param  slaveAddr
 *		I2C slave address of device.
 *  @param  memAddr
 *		The location in the memory to read from.
 *  @param  length
 *		Length of burst data.
 *  @param  data
 *		Pointer to block of data.
 *
 *  @return ML_SUCCESS == 0 if successful; a non-zero error code otherwise.
 */
tMLError MLSLSerialReadMem(void *sl_handle,
			   unsigned char slaveAddr,
			   unsigned short memAddr,
			   unsigned short length,
			   unsigned char *data);

/**
 *  @brief  used to write multiple bytes of data to the memory.
 *	    This should be sent by I2C or SPI.
 *
 *  @param  sl_handle
 *		a file handle to the serial device used for the communication.
 *  @param  slaveAddr
 *		I2C slave address of device.
 *  @param  memAddr
 *		The location in the memory to write to.
 *  @param  length
 *		Length of burst data.
 *  @param  data
 *		Pointer to block of data.
 *
 *  @return ML_SUCCESS == 0 if successful; a non-zero error code otherwise.
 */
tMLError MLSLSerialWriteMem(void *sl_handle,
			    unsigned char slaveAddr,
			    unsigned short memAddr,
			    unsigned short length,
			    unsigned char const *data);

/**
 *  @brief  used to read multiple bytes of data from the fifo.
 *	    This should be sent by I2C or SPI.
 *
 *  @param  sl_handle
 *		a file handle to the serial device used for the communication.
 *  @param  slaveAddr
 *		I2C slave address of device.
 *  @param  length
 *		Length of burst of data.
 *  @param  data
 *		Pointer to block of data.
 *
 *  @return ML_SUCCESS == 0 if successful; a non-zero error code otherwise.
 */
tMLError MLSLSerialReadFifo(void *sl_handle,
			    unsigned char slaveAddr,
			    unsigned short length,
			    unsigned char *data);

/**
 *  @brief  used to write multiple bytes of data to the fifo.
 *	    This should be sent by I2C or SPI.
 *
 *  @param  sl_handle
 *		a file handle to the serial device used for the communication.
 *  @param  slaveAddr
 *		I2C slave address of device.
 *  @param  length
 *		Length of burst of data.
 *  @param  data
 *		Pointer to block of data.
 *
 *  @return ML_SUCCESS == 0 if successful; a non-zero error code otherwise.
 */
tMLError MLSLSerialWriteFifo(void *sl_handle,
			     unsigned char slaveAddr,
			     unsigned short length,
			     unsigned char const *data);

/**
 *  @brief  used to get the configuration data.
 *	    Is called by the MPL to get the configuration data
 *	    used by the motion library.
 *	    This data would typically be saved in non-volatile memory.
 *
 *  @param  cfg
 *		Pointer to the configuration data.
 *  @param  len
 *		Length of the configuration data.
 *
 *  @return ML_SUCCESS if successful, a non-zero error code otherwise.
 */
tMLError MLSLReadCfg(unsigned char *cfg, unsigned int len);

/**
 *  @brief  used to save the configuration data.
 *	    Is called by the MPL to save the configuration data used by the
 *	    motion library.
 *	    This data would typically be saved in non-volatile memory.
 *
 *  @param  cfg
 *		Pointer to the configuration data.
 *  @param  len
 *		Length of the configuration data.
 *
 *  @return ML_SUCCESS if successful, a non-zero error code otherwise.
 */
tMLError MLSLWriteCfg(unsigned char *cfg, unsigned int len);

/**
 *  @brief  used to get the calibration data.
 *	    It is called by the MPL to get the calibration data used by the
 *	    motion library.
 *	    This data is typically be saved in non-volatile memory.
 *
 *  @param  cfg
 *		Pointer to the calibration data.
 *  @param  len
 *		Length of the calibration data.
 *
 *  @return ML_SUCCESS if successful, a non-zero error code otherwise.
 */
tMLError MLSLReadCal(unsigned char *cal, unsigned int  len);

/**
 *  @brief  used to save the calibration data.
 *	    It is called by the MPL to save the calibration data used by the
 *	    motion library.
 *	    This data is typically be saved in non-volatile memory.
 *
 *  @param  cfg
 *		Pointer to the calibration data.
 *  @param  len
 *		Length of the calibration data.
 *
 *  @return ML_SUCCESS if successful, a non-zero error code otherwise.
 */
tMLError MLSLWriteCal(unsigned char *cal, unsigned int len);

/**
 *  @brief  Get the calibration length from the storage.
 *  @param  len
 *		lenght to be returned
 *  @return ML_SUCCESS if successful, a non-zero error code otherwise.
 */
tMLError MLSLGetCalLength(unsigned int *len);



#ifdef __cplusplus
}
#endif

/**
 * @}
 */

#endif	/* __MLSL_H__ */
