/*
 $License:
    Copyright (C) 2010 InvenSense Corporation, All Rights Reserved.
 $
 */
/******************************************************************************
 *
 * $Id: umpl-setup-imu3k-evb.c 2011-04-15  phickey $
 *
 *****************************************************************************/

/**
 *  @defgroup UMPL
 *  @brief  uMPL Layer.
 *          umpl-setup-imu3k-evb provides a mechanism to perform IMU30X0 specific set-up.
 *
 *  @{
 *      @file   umpl-setup-imu3k-evb.c
 *      @brief  Performs set-up specific to platform.
**/

/* ------------------ */
/* - Include Files. - */
/* ------------------ */
#include <stdio.h>
#include <string.h>

#include "umpl-setup.h"

#include "ml.h"
#include "mldl.h"
#include "mlstates.h"
#include "mlFIFO.h"
#include "mpu.h"
#include "mldl_cfg.h"
#include "slave.h"
#include "tempComp.h"

#include "log.h"
#undef MPL_LOG_TAG
#define MPL_LOG_TAG "umpl-setup"

/* found in mldl.c */
extern struct ext_slave_descr gAccel;
extern struct mpu_platform_data gPdata;

/**
 *  @brief  umplPlatformSetup initializes the platform data structure to the required values.
 *            It sets the orientation matrix, the slave address of the gyro and the accel.
 *
 *
 *  @return ML_SUCCESS if successful, a non-zero error code otherwise.
 */

tMLError umplPlatformSetup(void)
{
    struct ext_slave_descr* accel = NULL;
    tMLError result;

    memset(&gPdata.orientation, 0, sizeof(gPdata.orientation));
    gPdata.orientation[0] = +1;
    gPdata.orientation[4] = +1;
    gPdata.orientation[8] = +1;
    gPdata.int_config = 0x30; // Interrupt: Latch, any read clears

    memset(&gAccel,0,sizeof(gAccel));

    accel = kxtf9_get_slave_descr();
    RANGE_FLOAT_TO_FIXEDPOINT(accel->range,2.f);

    memset(gPdata.accel.orientation, 0, sizeof(gPdata.accel.orientation));
    gPdata.accel.orientation[0] = +1;
    gPdata.accel.orientation[4] = +1;
    gPdata.accel.orientation[8] = +1;

    gPdata.accel.adapt_num = 0;
    gPdata.accel.bus = EXT_SLAVE_BUS_SECONDARY;
    gPdata.accel.get_slave_descr = kxtf9_get_slave_descr;
    gPdata.accel.address = ACCEL_SLAVEADDR_KXTF9;

    memcpy(&gAccel, accel, sizeof(*accel));

    return ML_SUCCESS;
}

void umplMotionCallback(unsigned short motionstate)
{
    if (motionstate == ML_MOTION)
        MPL_LOGI("umplMotionCallback: ML_MOTION\n");
    else if (motionstate == ML_NO_MOTION)
        MPL_LOGI("umplMotionCallback: ML_NO_MOTION\n");
}

/**
 *  @brief  umplDmpSetup is called by Called by umplStartMPU(). This function
 *            enables the required features such as turning on the bias trackers
 *            and temperature compensation. This function also enables the required
 *            type of data to be put in FIFO.
 *
 *
 *  @pre    To be run after MLDmpOpen(), before MLDmpStart().
 *
 *  @return ML_SUCCESS if successful, a non-zero error code otherwise.
 */
tMLError umplDmpSetup(void)
{
    tMLError result;

    // Setup Bias Trackers:
    /* MLSetBiasUpdateFunc
     * Known to work with the following arguments:
     * ML_BIAS_FROM_NO_MOTION
     * ML_BIAS_FROM_TEMPERATURE
     * ML_LEARN_BIAS_FROM_TEMPERATURE
     * ML_BIAS_FROM_GRAVITY
     *
     */

    result = MLSetBiasUpdateFunc( ML_BIAS_FROM_NO_MOTION
                                | ML_BIAS_FROM_TEMPERATURE
                                | ML_LEARN_BIAS_FROM_TEMPERATURE
                                /* Optional: */
                                //| ML_BIAS_FROM_GRAVITY
                                );
    if (result != ML_SUCCESS)
        MPL_LOGE("umpl MLSetBiasUpdateFunc returned %d\n",result);

    MLSetFifoInterrupt(TRUE);
    MLSetMotionInterrupt(TRUE);
    MLSetMotionCallback(&umplMotionCallback);

    // Setup FIFO:
    result = FIFOSendQuaternion(ML_32_BIT);
    if (result != ML_SUCCESS)
        MPL_LOGE("umpl FIFOSendQuaterion returned %d\n",result);

   result = FIFOSendGyro(ML_ALL,ML_32_BIT);
    if (result != ML_SUCCESS)
     MPL_LOGE("umpl FIFOSendGyro returned %d\n",result);

    result = FIFOSendAccel(ML_ALL,ML_32_BIT);
    if (result != ML_SUCCESS)
        MPL_LOGE("umpl FIFOSendAccel returned %d\n",result);

    result = MLSetFIFORate(10);
    if (result != ML_SUCCESS)
        MPL_LOGE("umpl MLSetFIFORate returned %d\n",result);

    result = MLEnableTempComp();
    if (result != ML_SUCCESS)
       MPL_LOGE("umpl MLEnableTempComp returned %d\n",result);
    return ML_SUCCESS;
}

