/*
 $License:
    Copyright (C) 2010 InvenSense Corporation, All Rights Reserved.
 $
 */
/******************************************************************************
 *
 * $Id: umpl-states.c 2011-04-15  phickey $
 *
 *****************************************************************************/

/**
 *  @defgroup UMPL
 *  @brief  uMPL Layer.
 *          umpl-states.c provides a mechanism to control the uMPL state machine.
 *            It controls the flow from one state to the next.
 *
 *  @{
 *      @file   umpl-states.c
 *      @brief  Controls the UMPL state machine.
**/

/* ------------------ */
/* - Include Files. - */
/* ------------------ */
#include "umpl-states.h"
#include "umpl-setup.h"

#include "log.h"
#undef MPL_LOG_TAG
#define MPL_LOG_TAG "umpl-states"

#include "ml.h"
#include "mlFIFO.h"
#include "mldl.h"
#include "ml_stored_data.h"

//#define UMPL_DISABLE_LOAD_CAL
//#define UMPL_DISABLE_STORE_CAL

/*
 * UMPL State Interface
 *
 */

static int umplState = UMPL_UNINIT;

/* --------------- */
/* -  Functions. - */
/* --------------- */

/**
 *  @brief  umplGetState returns the current state of UMPL.
 *
 *
 *  @return The current state of UMPL
 */
int umplGetState(void)
{
    return umplState;
}

/**
 *  @brief  umplSetState is called to set the umpl state to the required state.
 *
 *
 *  @param  data
 *              the state to which umpl must be set.
 *
 *  @return ML_SUCCESS if successful, a non-zero error code otherwise.
 */
static tMLError umplSetState(int s)
{
    /* illegal to set state to UMPL_UNINT */
    if ( s == UMPL_STOP || s == UMPL_LP_ACCEL || s == UMPL_RUN )
    {
        umplState = s;
        return ML_SUCCESS;
    }
    return ML_ERROR;
}

/**
 *  @brief  umplNotifyInterrupt is called to handle a recently occurred interrupt.
 *
 *
 *
 *  @pre    umplInit() and  umplStartAccel() or umplStartMPU()
 *          must have been called.
 *
 *  @param  data
 *              the occurred interrupt which must be handled
 *
 *  @return ML_SUCCESS if successful, a non-zero error code otherwise.
 */

tMLError umplNotifyInterrupt(unsigned char interrupt)
{
    if (umplState == UMPL_LP_ACCEL || umplState == UMPL_RUN)
    {
        return MLDLIntHandler(interrupt);
    }
    return ML_ERROR_SM_IMPROPER_STATE;
}

/**
 *  @brief  umplOnTimer calls MLUpdataData which updates all the realtime
 *            data from the motion algorithms.
 *
 *
 *
 *  @pre    umplInit() and  umplStartAccel() or umplStartMPU()
 *          must have been called.
 *
 *
 *  @return ML_SUCCESS if successful, a non-zero error code otherwise.
 */
tMLError umplOnTimer(void)
{
    if (umplState == UMPL_LP_ACCEL || umplState == UMPL_RUN)
    {
        tMLError result;

        result = MLUpdateData();

        return result;
    }
    return ML_ERROR_SM_IMPROPER_STATE;
}


/**
 *  @brief  umplInit is the entry point to uMPL and must be called before
 *            all other functions. It initializes the platform specific data.
 *            It also initializes the MPL state.
 *
 *
 *  @param  data
 *              A dummy port number. Could be NULL.
 *
 */


tMLError umplInit(unsigned char *port)
{
    if (umplState == UMPL_UNINIT)
    {
        tMLError result;

        result = umplPlatformSetup();
        if (result != ML_SUCCESS) return result;

        result = MLSerialOpen((char const *)port);
        if (result != ML_SUCCESS) return result;

        umplSetState(UMPL_STOP);
        return ML_SUCCESS;
    }
    return ML_ERROR_SM_IMPROPER_STATE;
}

/**
 *  @brief  umplStartMPU kicks starts the uMPL state machine. This function
 *            in turn calls the MPL functions like MLDmpOpen() and MLDmpStart() which starts
 *            the motion processing algorithms. This function also enables the required features
 *            such as turning on the bias trackers and temperature compensation.
 *            This function enables the required type of data to be put in FIFO.
 *
 *
 *
 *  @pre    umplInit() must have been called.
 *
 *  @return ML_SUCCESS if successful, a non-zero error code otherwise.
 */
tMLError umplStartMPU(void)
{
    if (umplState == UMPL_STOP) {   // trans STOP -> LP_ACCEL
        tMLError result;
        result = umplStartAccel();  /* PRESENTLY A STUB */
        if (result != ML_SUCCESS)
            return result;
    }

    if (umplState == UMPL_LP_ACCEL) {
        tMLError result;

        result = MLDmpOpen();
        if (result != ML_SUCCESS) return result;

        result = umplDmpSetup();
        if (result != ML_SUCCESS) return result;

        result = MLDmpStart();
        if (result != ML_SUCCESS) return result;

#ifndef UMPL_DISABLE_LOAD_CAL
        result = MLLoadCalibration();
        if (result != ML_SUCCESS)
            MPL_LOGE("MLLoadCalibration failed with %d in umplStartMPU\n", result);
#endif
        umplSetState(UMPL_RUN);
        return ML_SUCCESS;
    }
    return ML_ERROR_SM_IMPROPER_STATE;
}

/**
 *  @brief  umplStartAccel starts the accelerometer.
 *
 *
 *  @pre    umplInit() must have been called.
 *
 *  @return ML_SUCCESS if successful, a non-zero error code otherwise.
 */
tMLError umplStartAccel(void) /* PRESENTLY A STUB */
{
    if (umplState == UMPL_STOP) {
        // transition All off -> Accel Only
        umplSetState(UMPL_LP_ACCEL);
        return ML_SUCCESS;
    }
    return ML_ERROR_SM_IMPROPER_STATE;
}

/**
 *  @brief  umplStopMPU is called by umplStop.
 *
 *
 *  @pre    umplInit() and umplStartMPU() must have been called.
 *
 *  @return ML_SUCCESS if successful, a non-zero error code otherwise.
 */
tMLError umplStopMPU(void) /* PRESENTLY A STUB */
{
    if (umplState == UMPL_RUN) {
        // transition MPU On -> Accel Only
        umplSetState(UMPL_LP_ACCEL);
        return ML_SUCCESS;
    }
    return ML_ERROR_SM_IMPROPER_STATE;
}

/**
 *  @brief  umplStop stops the uMPL state machine. This function
 *            in turn calls the necessary MPL functions like MLDmpStop() and MLDmpClose()
 *            to stop the motion processing algorithms.
 *
 *
 *
 *  @pre    umplInit() and umplStartMPU() must have been called.
 *
 *  @return ML_SUCCESS if successful, a non-zero error code otherwise.
 */
tMLError umplStop(void)
{
    if (umplState == UMPL_RUN) {   // trans RUN -> LP_ACCEL
        tMLError result = umplStopMPU(); /* PRESENTLY A STUB */
        if (result != ML_SUCCESS)
            return result;
    }

    if (umplState == UMPL_LP_ACCEL) {
        tMLError result;

#ifndef UMPL_DISABLE_STORE_CAL
        result = MLStoreCalibration();
        if (result != ML_SUCCESS)
            MPL_LOGE("MLStoreCalibration failed with %d in umplStop\n",result);
#endif

        result = MLDmpStop();
        if (result != ML_SUCCESS) return result;

        result = MLDmpClose();
        if (result != ML_SUCCESS) return result;

        // transition Accel Only -> Accel Off

        umplSetState(UMPL_STOP);
        return ML_SUCCESS;
    }
    return ML_ERROR_SM_IMPROPER_STATE;
}

